package xyz.johnny.norntalk

import android.os.AsyncTask
import android.util.Log
import android.view.ContextMenu
import android.view.Menu
import android.view.MenuItem
import android.view.View
import android.widget.AdapterView
import android.widget.ListView
import kotlinx.android.synthetic.main.fragment_message.view.*
import kotlinx.android.synthetic.main.view_compose.*
import xyz.johnny.norntalk.R.id.action_mute
import xyz.johnny.norntalk.R.id.action_secure
import xyz.johnny.norntalk.database.NornDatabase
import xyz.johnny.norntalk.messages.*
import xyz.johnny.norntalk.ui_components.ComposeView
import xyz.johnny.norntalk.ui_components.DismissableActivity
import xyz.johnny.norntalk.ui_components.NornMessageAdapter
import java.lang.ref.WeakReference
import java.net.URLDecoder
import java.util.*

/**
 * Activité affichant les messages d'une conversation
 */
class ConversationActivity: ComposeView.SendActivity() {

    companion object {

        /**
         * conversations en premier plan
         */
        val foreground_activities = HashSet<Int>()

        class MessagesTask(activity: ConversationActivity) : AsyncTask<Void, Void, List<NornMessage>>() {

            val activity = WeakReference<ConversationActivity>(activity)

            override fun doInBackground(vararg params: Void): List<NornMessage> {
                val activity = this.activity.get()!!
                val db = NornDatabase.getNornDatabase(activity)
                return db.messageDao().getConversationMessages(activity.conversation.id).map {
                    NornMessage.fromDbMessage(it, activity)
                }
            }

            override fun onPostExecute(result: List<NornMessage>) {
                super.onPostExecute(result)
                try {
                    // créer la liste de messages
                    val activity = this.activity.get()!!
                    val container = activity.findViewById<ListView>(R.id.fragment_container)
                    activity.adapter = NornMessageAdapter(
                            result.toMutableList(),
                            activity,
                            activity.conversation.id
                    )

                    container.adapter = activity.adapter
                    activity.registerForContextMenu(container)
                }catch (e: Exception) {
                    e.stackTrace.forEach { println(it) }
                    throw e
                }
            }

        }

        class SendTask(activity: ConversationActivity, val text: String, private val medias: Array<NornMedia>) : AsyncTask<Void, Void, NornMessage?>() {

            val activity = WeakReference<ConversationActivity>(activity)

            override fun doInBackground(vararg params: Void): NornMessage? {
                val activity = this.activity.get()!!
                val msg = NornMessage(text, activity.conversation, Date(), activity)
                for (media in medias)
                    msg.medias.add(media)

                return if (msg.send(activity)) msg else null
            }

            override fun onPostExecute(nMessage: NornMessage?) {
                // transmettre le message à toutes les activités
                if (nMessage != null)
                    NornMessageDispatcher.send(nMessage)

                Log.d(this::class.java.simpleName, nMessage.toString())
            }

        }

    }

    override val layout = R.layout.activity_conversation

    /**
     * La liste des messages
     */
    private var adapter: NornMessageAdapter? = null
    /**
     * Identifiant de la conversation
     */
    private var _conversationId: Int? = null
    /**
     * Conversation associée à l'identifiant
     */
    private val conversation get() = NornConversation.getConversation(_conversationId!!, this)

    override fun initialize() {
        super.initialize()
        // lorsque l'utilisateur lance la conversation par l'application Contacts
        if (this.intent.action != null) {
            // trouver le contact dans l'Intent
            val num = URLDecoder.decode(this.intent.dataString.substringAfter(":"), "UTF-8")
            val contact = NornContact.getContact(num, this)

            // créer la conversation
            val conversation = NornConversation.getConversation(arrayOf(contact), this)
            this._conversationId = conversation.id

            // insérer le message s'il y en a un
            val body = this.intent.extras?.getString("sms_body")
            if (body != null) {
                composeInput.setText(body)
            } else if (conversation.secured) {
                composeInput.setHint(R.string.secured_message) //indiquer que la conversation est securisée
            }

            composeInput.requestFocus()
        } else {
            // trouver l'identifiant donné en argument à l'activité
            this._conversationId = this.intent.extras.getInt("conversationId")
            if (conversation.secured) {
                composeInput.setHint(R.string.secured_message) //indiquer que la conversation est securisée
            }
        }

        // lire les messages en arrière plan
        MessagesTask(this).executeOnExecutor(AsyncTask.THREAD_POOL_EXECUTOR)

        // ajouter le titre de la conversation
        this.title = this.conversation.name
    }

    override fun sendMessage(text: String, medias: Array<NornMedia>) {
        // envoyer le message en arrière plan
        SendTask(this, text, medias).executeOnExecutor(AsyncTask.THREAD_POOL_EXECUTOR)
    }

    override fun onCreateOptionsMenu(menu: Menu): Boolean {
        // créer le menu de la barre d'outils
        menuInflater.inflate(R.menu.menu_conversation, menu)

        // cocher la case conversation sécurisée si la conversation est sécurisée
        val secured = menu.findItem(action_secure)
        secured.isChecked = conversation.secured

        // cocher la case ingnorer la conversation si l'utilisateur a décidé de l'ignorer
        val muted = menu.findItem(action_mute)
        muted.isChecked = conversation.muted

        return true
    }

    /**
     * Créer la conversation sécurisée si la conversation n'est pas sécurisée
     * ou arrêter la conversation sécurisée si elle l'est déjà
     *
     * @param item le bouton du menu sur lequel l'utilisateur a appuyé
     */
    private fun toggleSecurity(item: MenuItem) {
        // changer l'état de la case
        val secured = !item.isChecked
        item.isChecked = secured

        if (secured) {
            this.conversation.createSecuredConversation(this)
            composeInput.setHint(R.string.secured_message)
        } else {
            this.conversation.stopSecuredConversation(this)
            composeInput.setHint(R.string.message)
        }
    }

    /**
     * Ignorer les futurs messages concernant cette conversation si elle n'est pas ignorée
     * ou recevoir de nouveau des notifications concernant cette conversation si elle est déjà ignorée
     *
     * @param item le bouton du menu sur lequel l'utilisateur a appuyé
     */
    private fun toggleMuted(item: MenuItem) {
        // changer l'état de la case
        val muted = !item.isChecked
        item.isChecked = muted

        this.conversation.setMuted(muted, this)
    }

    /**
     * Supprimer la conversation actuelle
     */
    private fun deleteConversation() {
        this.conversation.delete(this)
        this.finish()
    }

    override fun onOptionsItemSelected(item: MenuItem): Boolean {
        // répondre aux actions de l'utilisateur dans le menu de la barre d'outils
        return when (item.itemId) {
            R.id.action_secure -> { this.toggleSecurity(item); true }
            R.id.action_delete_conv -> { this.deleteConversation(); true }
            R.id.action_mute -> { this.toggleMuted(item); true }
            else -> super.onOptionsItemSelected(item)
        }

    }

    override fun onCreateContextMenu(menu: ContextMenu, v: View, menuInfo: ContextMenu.ContextMenuInfo) {
        super.onCreateContextMenu(menu, v, menuInfo)
        // créer le menu contextuel
        val items = this.resources.getStringArray(R.array.context_menu)
        for (i in items.indices)
            menu.add(Menu.NONE, i, i, items[i])
    }

    override fun onContextItemSelected(item: MenuItem): Boolean {
        super.onContextItemSelected(item)

        // répondre aux actions de l'utilisateur sur le menu contextuel
        val items = this.resources.getStringArray(R.array.context_menu)
        val action = items[item.itemId]

        // trouver le message concerné par l'action
        val info = item.menuInfo as AdapterView.AdapterContextMenuInfo
        val message = info.targetView.messageText.message

        when(action) {
            getString(R.string.delete) -> message.delete(this)
        }

        return true
    }

    override fun onDestroy() {
        super.onDestroy()
        // désabonner l'adaptateur lorsque l'activité est détruite
        if (this.adapter != null)
            NornMessageDispatcher.unsubscribe(this.adapter!!)
    }

    override fun onResume() {
        super.onResume()
        DismissableActivity.dismiss()
        // ne plus montrer les notifications concernant la conversation actuelle lorsque l'activité
        // est au premier plan
        foreground_activities.add(this.conversation.id)
        this.conversation.clear(this)
    }

    override fun onPause() {
        super.onPause()
        // montrer de nouveau les notifications concernant la conversation actuelle lorsque
        // l'utilisateur quitte l'écran
        foreground_activities.remove(this.conversation.id)
    }

}