package xyz.johnny.norntalk.database

import android.arch.persistence.db.SupportSQLiteDatabase
import android.arch.persistence.room.Database
import android.arch.persistence.room.Room
import android.arch.persistence.room.RoomDatabase
import android.arch.persistence.room.migration.Migration
import android.content.Context
import android.os.AsyncTask
import android.os.AsyncTask.THREAD_POOL_EXECUTOR
import xyz.johnny.norntalk.database.dao.*
import xyz.johnny.norntalk.database.entities.*
import java.lang.ref.WeakReference

/**
 * Base de données de Norn Talk
 */
@Database(entities = arrayOf(Message::class, Member::class, Media::class,
          Conversation::class, Contact::class), version=6, exportSchema=false)
abstract class NornDatabase : RoomDatabase() {

    companion object {
        /**
         * Nom de la base de données
         */
        private val DB_NAME = "norn_db"
        /**
         * Instance unique de la base de données
         */
        private var instance: NornDatabase? = null

        /**
         * Migration de la base de données de la version 2 à la version 3.
         * Rajoute le champs read de l'entité Message.
         */
        private val migrations = arrayOf(
            object : Migration(2, 3) {

                override fun migrate(database: SupportSQLiteDatabase) {
                    database.execSQL("alter table Message add column read integer not null default 0")
                }

            },
            object: Migration(4, 5) {

                override fun migrate(database: SupportSQLiteDatabase) {
                    database.execSQL("alter table Media add column name text not null default ''")
                    database.execSQL("alter table Media add column position integer not null default 0")
                }

            }
        )

        class BuildDBTask : AsyncTask<Context, Void, Unit>() {

            override fun doInBackground(vararg params: Context) {
                val dbBuilder = Room.databaseBuilder(
                        params[0].applicationContext, NornDatabase::class.java, DB_NAME
                ).fallbackToDestructiveMigration()

                for (migration in migrations) {
                    dbBuilder.addMigrations(migration)
                }

                instance = dbBuilder.build()
            }

        }

        /**
         * Tâche permettant de créer la base de données
         */
        private val task by lazy { BuildDBTask() }

        /**
         * Fonction permettant d'accéder à l'instance de la base de données
         *
         * @param context Contexte courant
         */
        fun getNornDatabase(context: Context): NornDatabase {
            if (instance == null) {
                if (this.task.status == AsyncTask.Status.PENDING)
                    // lancer la création de la base de données dans un autre thread
                    this.task.executeOnExecutor(THREAD_POOL_EXECUTOR, context)

                if (this.task.status == AsyncTask.Status.RUNNING)
                    // attendre que la base de données soit créée
                    this.task.get()
            }

            return instance!!
        }

    }

    /**
     * Classe permettant de facilement utiliser la base de données dans un thread différent
     *
     * @property context Contexte courant
     */
    class Helper(val context: Context) {

        companion object {
            class RunTask<T>(context: Context, val function: (db: NornDatabase) -> T) : AsyncTask<Void, Void, T>() {

                val context = WeakReference<Context>(context)

                override fun doInBackground(vararg params: Void?): T {
                    return function(getNornDatabase(this.context.get()!!))
                }

            }
        }

        /**
         * Execute la fonction donnée dans une tâche effectuée en parallèle
         *
         * @param T Type de sortie de la tâche
         * @param function Fonction acceptant un argument et retournant un objet de type T
         * @return La tâche correspondante à l'opération, en cours d'execution
         */
        fun <T> run(function: (db: NornDatabase) -> T): AsyncTask<Void, Void, T> {
            return RunTask<T>(context, function).executeOnExecutor(AsyncTask.THREAD_POOL_EXECUTOR)
        }

    }

    /**
     * Retourne une instance de [MessageDao] permettant d'accéder aux messages ([Message]) contenus
     * dans la base de données
     *
     * @return Instance de [MessageDao]
     */
    abstract fun messageDao(): MessageDao

    /**
     * Retourne une instance de [MemberDao] permettant d'accéder aux membres ([Member]) contenus
     * dans la base de données
     *
     * @return Instance de [MemberDao]
     */
    abstract fun memberDao(): MemberDao

    /**
     * Retourne une instance de [MediaDao] permettant d'accéder aux médias ([Media]) contenus
     * dans la base de données
     *
     * @return Instance de [MediaDao]
     */
    abstract fun mediaDao(): MediaDao

    /**
     * Retourne une instance de [ConversationDao] permettant d'accéder aux conversations
     * ([Conversation]) contenus dans la base de données
     *
     * @return Instance de [ConversationDao]
     */
    abstract fun conversationDao(): ConversationDao

    /**
     * Retourne une instance de [ContactDao] permettant d'accéder aux contacts ([Contact]) contenus
     * dans la base de données
     *
     * @return Instance de [ContactDao]
     */
    abstract fun contactDao(): ContactDao

}