package xyz.johnny.norntalk.database.dao

import android.arch.persistence.room.*
import xyz.johnny.norntalk.database.entities.Conversation

/**
 * Opérations associées à l'entité [Conversation]
 */
@Dao
interface ConversationDao {

    /**
     * Retourne l'ensemble des conversations dans la base de données
     *
     * @return Ensemble des conversations
     */
    @Query("select * from Conversation")
    fun getAllConversations(): List<Conversation>

    /**
     * Retourne la conversation à une ligne données, utile après avoir insérer une conversation
     * pour trouvé l'id qui lui a été associé.
     *
     * @param rowid Ligne retourné par [insertConversation]
     * @return Conversation correspondante
     */
    @Query("select * from Conversation where rowid = :rowid")
    fun getConversationFromRowId(rowid: Long): Conversation

    /**
     * Retourne la conversation correspondant à un identifiant donné
     *
     * @param id Identifiant de la conversation
     * @return conversation correspondante
     */
    @Query("select * from Conversation where id = :id")
    fun getConversationFromId(id: Int): Conversation

    /**
     * Retourne la conversation regroupant tous les contacts donnés (identifiés par leur numéro de
     * téléphone)
     *
     * @param numbers Numéros de téléphone des contacts
     * @param len Nombre de contacts dans la conversation
     * @return Conversation correspondante ou null si elle n'existe pas
     */
    @Query("""
        select id, notification, secured
        from Conversation, Member
        where Conversation.id = Member.conversationId
          and Member.contactNumber in (:numbers)
        group by id, notification
        having count(*) == :len
           and conversationId not in (select distinct conversationId
                                      from Member
                                      where Member.contactNumber not in (:numbers))
        """)
    fun getConversation(numbers: Array<String>, len: Int): Conversation?

    /**
     * Insére une conversation donné
     *
     * @param conversation Conversation à insérer
     * @return Ligne où la conversation a été insérée
     */
    @Insert
    fun insertConversation(conversation: Conversation): Long

    /**
     * Mettre à jour une conversation
     *
     * @param conversation Conversation à mettre à jour
     */
    @Update
    fun updateConversation(conversation: Conversation)

    /**
     * Supprimer une conversation
     *
     * @param conversation Conversation à supprimer
     */
    @Delete
    fun deleteConversation(conversation: Conversation)

}