package xyz.johnny.norntalk.database.dao

import android.arch.persistence.room.*
import xyz.johnny.norntalk.database.entities.Message

/**
 * Opérations associées à l'entité [Message]
 */
@Dao
interface MessageDao {

    /**
     * Retourne l'ensemble des messages dans la base de données
     *
     * @return Ensemble des messages
     */
    @Query("select * from Message")
    fun getAllMessages(): List<Message>

    /**
     * Retourne le dernier message de toutes les conversations existantes
     *
     * @return dernier message de chaque conversation
     */
    @Query("""
        select *
        from Message
        where id in (
            select id
            from Message
            group by conversationId
            having date = max(date)
        )
        order by date desc
    """)
    fun getLastMessages(): List<Message>

    /**
     * Retourne le dernier message d'une conversation donnée
     *
     * @param convId Identifiant de la conversation
     * @return Dernier message de la conversation
     */
    @Query("""
        select * from Message
        where conversationId = :convId
          and id in (
            select id
            from Message
            group by conversationId
            having date = max(date)
        )
        order by date desc
    """)
    fun getLastMessage(convId: Int): Message?

    /**
     * Retourne tous les messages d'une conversation
     *
     * @param conversationId Identifiant de la conversation
     * @return Ensemble des messages de la conversation
     */
    @Query("""
        select *
        from Message
        where conversationId = :conversationId
        order by date asc
    """)
    fun getConversationMessages(conversationId: Int): List<Message>

    /**
     * Retourne le message à une ligne donnée. Utile après avoir insérer le message à l'aide de
     * la méthode [insertMessage]
     *
     * @param rowId Ligne donnée
     * @return Message à la ligne donnée
     */
    @Query("select * from Message where rowid = :rowId")
    fun getMessageByRowId(rowId: Long): Message

    /**
     * Marque les messages d'une conversation comme lus
     *
     * @param conversationId Identifiant de la conversation
     */
    @Query("update Message set read = 1 where conversationId = :conversationId")
    fun readConversation(conversationId: Int)

    /**
     * Insère un message dans la base donnée
     *
     * @param message Message à insérer
     * @return Ligne à laquelle le message a été inséré
     */
    @Insert
    fun insertMessage(message: Message): Long

    /**
     * Mettre à jour un message donné
     *
     * @param message Message à mettre à jour
     */
    @Update
    fun updateMessage(message: Message)

    /**
     * Supprimer un message donné
     *
     * @param message Message à supprimer
     */
    @Delete
    fun deleteMessage(message: Message)

}