package xyz.johnny.norntalk.messages

import android.content.ContentUris
import android.content.Context
import android.database.Cursor
import android.graphics.drawable.Drawable
import android.net.Uri
import android.provider.ContactsContract
import android.support.v4.graphics.drawable.RoundedBitmapDrawableFactory
import android.telephony.PhoneNumberUtils
import android.telephony.TelephonyManager
import xyz.johnny.norntalk.R
import xyz.johnny.norntalk.database.NornDatabase
import xyz.johnny.norntalk.database.entities.Contact

/**
 * Classe liée aux contacts [Contact].
 * Effectue des opérations permettant de simplifier l'utilisation des données de [Contact].
 *
 * @property number Numéro de téléphone du contact
 * @param context Contexte courant
 * @constructor Crée un contact à partir d'un numéro de téléphone
 */
class NornContact private constructor(val number: String, context: Context) {

    companion object {

        /**
         * Association numéro de téléphone/[NornContact].
         */
        private val contacts = HashMap<String, NornContact>()

        /**
         * Formate le numéro de téléphone
         *
         * @param number Numéro de téléphone
         * @param context Contexte courant
         * @return Numéro de téléphone formaté
         */
        fun format(number: String, context: Context): String {
            val telMan = context.getSystemService(Context.TELEPHONY_SERVICE) as TelephonyManager
            val numberE164 = PhoneNumberUtils.formatNumberToE164(number, telMan.networkCountryIso.toUpperCase())

            return PhoneNumberUtils.normalizeNumber(numberE164)
        }

        /**
         * Retourne le contact associé à un numéro de téléphone
         *
         * @param number Numéro de téléphone
         * @param context Contexte courant
         * @return Contact [NornContact]
         */
        @Synchronized
        fun getContact(number: String, context: Context): NornContact {
            val num = format(number, context)

            val contact = contacts[num]
            if (contact != null)
                return contact

            return NornContact(num, context)
        }

        /**
         * Retourne le contact correspondant à un contact [Contact] donné
         *
         * @param dbContact Contact
         * @param context Contexte courant
         * @return Contact [NornContact]
         */
        @Synchronized
        fun getContact(dbContact: Contact, context: Context): NornContact {
            val contact = contacts[dbContact.number]
            if (contact != null)
                return contact

            return NornContact(dbContact.number, context)
        }

    }

    /**
     * Nom du contact
     */
    val name: String
    /**
     * Identifiant du contact
     */
    val id: Long?
    /**
     * URI du contact
     */
    val uri: Uri? get() {
        if (this.id == null)
            return null

        return ContentUris.withAppendedId(ContactsContract.Contacts.CONTENT_URI, this.id)
    }
    /**
     * Avatar du contact
     */
    val avatar: Drawable?
    /**
     * Contact dans la base de données
     */
    val db: Contact
    /**
     * Clé publique de l'utilisateur
     */
    val pubKey: ByteArray?
        get() = db.pubkey

    init {
        // conserver l'instance
        contacts.put(number, this)

        // retrouver ou créer l'entrée conrespondante dans la base de données
        val task = NornDatabase.Helper(context).run<Contact> { db ->
            var contact: Contact? = db.contactDao().getContact(number)

            if (contact == null) {
                contact = Contact(number, null)
                db.contactDao().insertContact(contact)
            }

            contact
        }

        // trouver le nom de l'utilisateur
        val contact = getContact(context)
        val name = contact.get("name") as String?
        this.name = if (name == null) this.number else name

        // conserver l'id de l'utilisateur
        this.id = contact.get("id") as Long?

        // trouver l'avatar de l'utilisateur
        this.avatar = this.getAvatar(context) ?: context.getDrawable(R.drawable.ic_account_circle)

        this.db = task.get()
    }

    /**
     * Retourne l'avatar associé au contact, ou null si l'avatar est introuvable
     *
     * @param context Contexte courant
     * @return Avatar
     */
    private fun getAvatar(context: Context): Drawable? {
        // retourner null si l'id est introuvable
        if (this.id == null)
            return null

        // lire l'image
        val uri = ContentUris.withAppendedId(ContactsContract.Contacts.CONTENT_URI, this.id)
        val stream = ContactsContract.Contacts.openContactPhotoInputStream(context.contentResolver,
                                                                           uri, false)

        // retourner null si l'image est introuvable
        if (stream == null)
            return null

        // retourner un avatar arrondi
        val drawable = RoundedBitmapDrawableFactory.create(context.resources, stream)
        drawable.cornerRadius = Math.max(drawable.bitmap!!.width, drawable.bitmap!!.height) / 2f

        return drawable
    }

    /**
     * Retourne l'URI permettant d'accéder aux données du contact
     *
     * @return URI
     */
    private fun getPhoneLookupUri(): Uri {
        return Uri.withAppendedPath(ContactsContract.PhoneLookup.CONTENT_FILTER_URI,
                Uri.encode(this.number))
    }

    /**
     * Effectue la requête permettant de retrouver le nom et l'identifiant du contact
     *
     * @param context Contexte courant
     * @return le nom et l'identifiant du contact
     */
    private fun getContact(context: Context): Map<String, Any?> {
        val projection = arrayOf(ContactsContract.PhoneLookup.DISPLAY_NAME,
                                 ContactsContract.PhoneLookup._ID)

        var cursor : Cursor? = null
        var name: String? = null
        var id: Long? = null

        try {
            cursor = context.contentResolver.query(this.getPhoneLookupUri(), projection, null, null, null)
            // récupérer le nom et l'identifiant du contact

            if (cursor != null) {
                if (cursor.moveToFirst()) {
                    name = cursor.getString(cursor.getColumnIndexOrThrow(ContactsContract.PhoneLookup.DISPLAY_NAME))
                    id = cursor.getLong(cursor.getColumnIndexOrThrow(ContactsContract.PhoneLookup._ID))
                }
            }

        } catch (e: IllegalArgumentException) {
            throw IllegalArgumentException(context.getString(R.string.invalid_number))
        } finally {
            if (cursor != null)
                cursor.close()
        }

        // retourner le nom et l'identifiant dans un dictionnaire
        return mapOf("name" to name, "id" to id)
    }

    /**
     * Mets à jour la clé publique du contact
     *
     * @param publicKey Clé publique
     * @param context Contexte courant
     */
    fun setPublicKey(publicKey: ByteArray, context: Context) {
        this.db.pubkey = publicKey

        NornDatabase.Helper(context).run { db ->
            db.contactDao().updateContact(this.db)
        }
    }

    override fun equals(other: Any?): Boolean {
        if (this === other) return true
        if (javaClass != other?.javaClass) return false

        other as NornContact

        if (number != other.number) return false
        if (uri != other.uri) return false
        if (name != other.name) return false
        if (id != other.id) return false
        if (db != other.db) return false

        return true
    }

    override fun hashCode(): Int {
        return this.number.hashCode()
    }

    override fun toString(): String {
        return "NornContact(number='$number', name=$name, id=$id)"
    }

}