package xyz.johnny.norntalk.messages

import android.content.Context
import android.util.Log
import android.util.SparseArray
import org.bouncycastle.util.encoders.Hex
import xyz.johnny.norntalk.R
import xyz.johnny.norntalk.database.NornDatabase
import xyz.johnny.norntalk.database.entities.Conversation
import xyz.johnny.norntalk.database.entities.Member
import xyz.johnny.norntalk.security.ChaCha
import xyz.johnny.norntalk.security.Curve
import xyz.johnny.norntalk.ui_components.NornListAdapter
import java.util.*


/**
 * Classe liée aux conversations [Conversation].
 * Permet d'effectuer les principales opérations liées aux conversations
 *
 * @property members Membres de la conversation
 * @property db Entrée [Conversation] dans la base de donnée
 * @param context Contexte courant
 * @constructor Crée une conversation à partir des membres et de son entrée dans la base de données
 */
class NornConversation private constructor(val members: Array<NornContact>, val db: Conversation, context: Context) {

    companion object {

        /**
         * Association Identifiant conversation/[NornConversation]
         */
        private val conversations = SparseArray<NornConversation>()
        /**
         * Association Membres [NornContact]/[NornConversation]
         */
        private val conversationsFromMembers = SparseArray<NornConversation>()

        /**
         * Retourne la conversation correspondant à l'identifiant donné
         *
         * @param conversationId Identifiant de la conversation
         * @param context Contexte courant
         * @return Conversation [NornConversation]
         */
        @Synchronized
        fun getConversation(conversationId: Int, context: Context): NornConversation {
            val conversation = conversations[conversationId]
            if (conversation != null)
                return conversation

            return NornConversation(conversationId, context)
        }

        /**
         * Retourne la conversation [NornConversation] associée à un groupe de contacts
         *
         * @param members groupe de contacts
         * @param context Contexte courant
         * @return [NornConversation] associée à ce groupe de contacts
         */
        @Synchronized
        fun getConversation(members: Array<NornContact>, context: Context): NornConversation {
            val conversation = conversationsFromMembers[members.sumBy { it.hashCode() }]
            if (conversation != null)
                return conversation

            return NornConversation(members, context)
        }

        /**
         * Retourne l'entrée [Conversation] de la base de données associée à un groupe de contacts
         *
         * @param members Groupe de contacts
         * @param context Contexte courant
         * @return [Conversation] associée à ce groupe de contact
         */
        fun getDbConversation(members: Array<NornContact>, context: Context): Conversation {
            val task = NornDatabase.Helper(context).run<Conversation> { db ->
                var conv = db.conversationDao().getConversation(
                        members.map { it.number }.toTypedArray(), members.size
                )
                if (conv == null)
                    conv = createConversation(members, db)

                conv
            }

            return task.get()
        }

        /**
         * Retourne la [Conversation] associée à un identifiant de conversation
         *
         * @param conversationId Identifiant de conversation
         * @param context Contexte courant
         * @return [Conversation] associée à l'identifiant
         */
        fun getDbConversation(conversationId: Int, context: Context): Conversation {
            val task = NornDatabase.Helper(context).run { db ->
                db.conversationDao().getConversationFromId(conversationId)
            }

            return task.get()
        }

        /**
         * Crée une conversation à partir d'un groupe de contact
         *
         * @param members Groupe de contacts
         * @param db Base de données
         * @return La [Conversation] crée
         */
        private fun createConversation(members: Array<NornContact>, db: NornDatabase): Conversation {
            var conversation = Conversation(true, false)
            val rowid = db.conversationDao().insertConversation(conversation)

            conversation = db.conversationDao().getConversationFromRowId(rowid)

            for (contact in members) {
                val member = Member(conversation.id, contact.number)

                db.memberDao().insertMember(member)
            }

            return conversation
        }

        /**
         * Retourne les membres d'une conversation associée à l'identifiant donné
         *
         * @param conversationId Identifiant de la conversation
         * @param context Contexte courant
         * @return groupe de contact [NornContact]
         */
        fun getMembers(conversationId: Int, context: Context): Array<NornContact> {
            val members = NornDatabase.Helper(context).run { db ->
                db.contactDao().getContactsFromConversation(conversationId)
            }.get()

            return members.map { NornContact.getContact(it, context) }.toTypedArray()
        }

    }

    /**
     * Ensemble des message de la conversation
     */
    val messages = HashMap<Long, NornMessage>()

    /**
     * Vrai si la conversation est sécurisée, Faux sinon
     */
    val secured get() = db.secured

    /**
     * Vrai si l'utilisateur ne souhaite pas recevoir de notification provenant de cette conversation,
     * Faux sinon
     */
    val muted get() = !db.notification

    /**
     * Identifiant du message
     */
    val id get() = db.id

    /**
     * Nom de la conversation
     */
    val name = this.members.joinToString(", ") { it.name }

    /**
     * Instance de [NornNotification] associée à la conversation
     */
    val notification = NornNotification.getNotification(this, context)

    init {
        conversations.put(this.id, this)
        conversationsFromMembers.put(Arrays.hashCode(this.members), this)
    }

    /**
     * Création de la conversation à partir d'un groupe de contacts
     *
     * @param members Groupe de contacts
     * @param context Contexte courant
     */
    private constructor(members: Array<NornContact>, context: Context) :
            this(members, getDbConversation(members, context), context)

    /**
     * Création de la conversation à partir d'un identifiant de conversation
     *
     * @param conversationId Identifiant de conversation
     * @param context Contexte courant
     */
    private constructor(conversationId: Int, context: Context) :
            this(getMembers(conversationId, context), getDbConversation(conversationId, context), context)

    /**
     * Envoie le message permettant de démarrer une conversation sécurisée
     *
     * @param context Contexte courant
     * @param token Token inséré au début du message, prend la valeur "ke:" par défaut
     */
    fun createSecuredConversation(context: Context, token: String = context.getString(R.string.key_exchange_prefix), tempkey: ByteArray? = null) {
        Log.d(this::class.java.simpleName, "creating secured conversationId")

        val curve = Curve.getCurve(context)

        val pubkey = if (tempkey == null) curve.publicKey
                     else ChaCha(tempkey).encrypt(curve.publicKey)

        val text = token + Hex.toHexString(pubkey)
        Log.d(this::class.java.simpleName, "token:" + text)
        val message = NornMessage(text, this, Date(), context)

        message.send(context, false)
    }

    /**
     * Envoie le message permettant d'arrêter une conversation sécurisée
     *
     * @param context Contexte courant
     */
    fun stopSecuredConversation(context: Context) {
        Log.d(this::class.java.simpleName, "disabling secured conversationId")
        this.setSecured(false, context)
        val message = NornMessage(context.getString(R.string.end_secret_conversation), this, Date(), context)
        message.send(context, false)
    }

    /**
     * Met à jour l'attribut secured dans la base de données
     *
     * @param secured La nouvelle valeur de l'attribut
     * @param context Contexte courant
     */
    fun setSecured(secured: Boolean, context: Context) {
        if (secured == this.secured)
            return

        this.db.secured = secured
        Log.d(this::class.java.simpleName, "secured: " + this.secured.toString())

        NornDatabase.Helper(context.applicationContext).run { db ->
            db.conversationDao().updateConversation(this.db)
        }
    }

    /**
     * Met à jour l'attribut muted dans la base de données
     *
     * @param muted La nouvelle valeur de l'attribut
     * @param context Contexte courant
     */
    fun setMuted(muted: Boolean, context: Context) {
        if (muted == this.muted)
            return

        this.db.notification = !muted

        NornDatabase.Helper(context.applicationContext).run { db ->
            db.conversationDao().updateConversation(this.db)
        }
    }

    /**
     * Supprimer la conversation
     *
     * @param context Contexte courant
     */
    fun delete(context: Context) {
        NornDatabase.Helper(context).run { db ->
            db.conversationDao().deleteConversation(this.db)
        }

        NornMessageDispatcher.delete(this.id)
    }

    override fun toString(): String {
        return "NornConversation(id=${db.id}, secured=${this.secured}, members=${Arrays.toString(members)})"
    }

    /**
     * Marquer les message de la conversation comme lus
     *
     * @param context Contexte courant
     */
    fun clear(context: Context) {
        for (message in this.messages.values) {
            if (!message.read) {
                message.dbMessage.read = true
            }
        }
        NornDatabase.Helper(context).run { db ->
            db.messageDao().readConversation(this.id)
        }

        val mainAdapter = NornMessageDispatcher.mainAdapter
        if (mainAdapter != null) {
            for (adapter in mainAdapter) {
                if (adapter is NornListAdapter)
                    adapter.update.sendEmptyMessage(0)
            }
        }

        this.notification.clear()
    }

}