/*
 * Copyright (C) 2015 Jacob Klinker
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package xyz.johnny.norntalk.messages

import android.content.BroadcastReceiver
import android.content.Context
import android.content.Intent
import android.net.Uri
import android.os.AsyncTask
import android.os.Parcelable
import android.provider.Telephony
import android.telephony.SmsManager
import android.util.Log
import com.android.mms.service_alt.MmsConfig
import com.android.mms.transaction.DownloadManager
import com.android.mms.transaction.HttpUtils
import com.android.mms.transaction.TransactionSettings
import com.android.mms.util.SendingProgressTokenManager
import com.google.android.mms.InvalidHeaderValueException
import com.google.android.mms.MmsException
import com.google.android.mms.pdu_alt.*
import com.google.android.mms.pdu_alt.PduHeaders.STATUS_RETRIEVED
import com.google.android.mms.util_alt.SqliteWrapper
import com.klinker.android.send_message.Utils
import java.io.*
import java.lang.ref.WeakReference
import java.nio.charset.Charset
import java.util.*
import java.util.concurrent.Executors

class NornMMSReceiver : BroadcastReceiver() {

    // Override this and provide the MMSC to send the ACK to.
    // some carriers will download duplicate MMS messages without this ACK. When using the
    // system sending method, apparently Google does not do this for us. Not sure why.
    // You might have to have users manually enter their APN settings if you cannot get them
    // from the system somehow.
    val mmscInfoForReceptionAck: MmscInformation?
        get() = null

    override fun onReceive(context: Context, intent: Intent) {
        Log.d(TAG, "MMS has finished downloading, persisting it to the database")

        val path = intent.getStringExtra(EXTRA_FILE_PATH)
        Log.d(TAG, path)

        try {
            val mDownloadFile = File(path)
            val nBytes = mDownloadFile.length().toInt()
            FileInputStream(mDownloadFile).use { reader ->
                val response = ByteArray(nBytes)
                reader.read(response, 0, nBytes)
                val tasks = getNotificationTask(context, intent, response)

                val pdu = PduParser(response).parse() as RetrieveConf

                val message = this.createMessage(context, pdu)
                message.insertMessage(context)
                NornMessageDispatcher.send(message)

                Log.v(TAG, "response saved successfully")
                Log.v(TAG, "response length: " + response.size)
                mDownloadFile.delete()

                if (tasks != null) {
                    Log.v(TAG, "running the common async notifier for download")
                    for (task in tasks)
                        task.executeOnExecutor(RECEIVE_NOTIFICATION_EXECUTOR)
                }
            }

        } catch (e: FileNotFoundException) {
            val errorMessage = "MMS received, file not found exception"
            Log.e(TAG, errorMessage, e)
        } catch (e: IOException) {
            val errorMessage = "MMS received, io exception"
            Log.e(TAG, errorMessage, e)
        }

        handleHttpError(context, intent)
        DownloadManager.finishDownload(intent.getStringExtra(EXTRA_LOCATION_URL))
    }

    fun createMessage(context: Context, pdu: RetrieveConf): NornMessage {
        val medias = ArrayList<NornMedia>()
        var text = ""

        for (i in 0..(pdu.body.partsNum - 1)) {
            val part = pdu.body.getPart(i)
            val mimeType = part.contentType!!.toString(Charset.defaultCharset())

            if (mimeType.startsWith("image")) {
                val uri = this.writeFile(context, part.data)
                medias.add(NornMedia(context, mimeType, uri))
            } else if (mimeType == "text/plain")
                text = part.data.toString(Charset.defaultCharset())
        }

        val addresses = setOf(pdu.from.string).plus(pdu.to.map { it.string }.toTypedArray())
        val members = addresses.map { NornContact.getContact(it, context) }.toTypedArray()
        val conversation = NornConversation.getConversation(members, context)
        val message = NornMessage(text, pdu.from.string, conversation, Date(), context)
        for (media in medias)
            message.medias.add(media)

        return message
    }

    fun writeFile(context: Context, data: ByteArray): String {
        val fileName = "media." + System.currentTimeMillis() + "_" + Math.abs(Random().nextInt()).toString() + ".dat"
        val file = File(context.filesDir, fileName)

        FileOutputStream(file).use { writer ->
            writer.write(data)
        }

        return fileName
    }

    private fun handleHttpError(context: Context, intent: Intent) {
        val httpError = intent.getIntExtra(SmsManager.EXTRA_MMS_HTTP_STATUS, 0)
        if (httpError == 404 || httpError == 400) {
            // Delete the corresponding NotificationInd
            SqliteWrapper.delete(context,
                    context.contentResolver,
                    Telephony.Mms.CONTENT_URI,
                    LOCATION_SELECTION,
                    arrayOf(Integer.toString(PduHeaders.MESSAGE_TYPE_NOTIFICATION_IND), intent.getStringExtra(EXTRA_LOCATION_URL)))
        }
    }

    private abstract class CommonAsyncTask internal constructor(context: Context, protected val mTransactionSettings: TransactionSettings, internal val mNotificationInd: NotificationInd) : AsyncTask<Void, Void, Void>() {
        private val context = WeakReference<Context>(context)
        internal val mContentLocation: String

        init {
            mContentLocation = String(mNotificationInd.contentLocation)
        }

        /**
         * A common method to send a PDU to MMSC.
         *
         * @param pdu A byte array which contains the data of the PDU.
         * @param mmscUrl Url of the recipient MMSC.
         * @return A byte array which contains the response data.
         * If an HTTP error code is returned, an IOException will be thrown.
         * @throws java.io.IOException if any error occurred on network interface or
         * an HTTP error code(>=400) returned from the server.
         * @throws com.google.android.mms.MmsException if pdu is null.
         */
        @Throws(IOException::class, MmsException::class)
        internal fun sendPdu(pdu: ByteArray, mmscUrl: String): ByteArray {
            return sendPdu(SendingProgressTokenManager.NO_TOKEN, pdu, mmscUrl)
        }

        /**
         * A common method to send a PDU to MMSC.
         *
         * @param pdu A byte array which contains the data of the PDU.
         * @return A byte array which contains the response data.
         * If an HTTP error code is returned, an IOException will be thrown.
         * @throws java.io.IOException if any error occurred on network interface or
         * an HTTP error code(>=400) returned from the server.
         * @throws com.google.android.mms.MmsException if pdu is null.
         */
        @Throws(IOException::class, MmsException::class)
        internal fun sendPdu(pdu: ByteArray): ByteArray {
            return sendPdu(SendingProgressTokenManager.NO_TOKEN, pdu,
                    mTransactionSettings.mmscUrl)
        }

        /**
         * A common method to send a PDU to MMSC.
         *
         * @param token The token to identify the sending progress.
         * @param pdu A byte array which contains the data of the PDU.
         * @param mmscUrl Url of the recipient MMSC.
         * @return A byte array which contains the response data.
         * If an HTTP error code is returned, an IOException will be thrown.
         * @throws java.io.IOException if any error occurred on network interface or
         * an HTTP error code(>=400) returned from the server.
         * @throws com.google.android.mms.MmsException if pdu is null.
         */
        @Throws(IOException::class, MmsException::class)
        private fun sendPdu(token: Long, pdu: ByteArray?,
                            mmscUrl: String?): ByteArray {
            if (pdu == null) {
                throw MmsException()
            }

            if (mmscUrl == null) {
                throw IOException("Cannot establish route: mmscUrl is null")
            }

            val mContext = this.context.get()

            if (com.android.mms.transaction.Transaction.useWifi(mContext)) {
                return HttpUtils.httpConnection(
                        mContext, token,
                        mmscUrl,
                        pdu, HttpUtils.HTTP_POST_METHOD,
                        false, null, 0)
            }

            Utils.ensureRouteToHost(mContext, mmscUrl, mTransactionSettings.proxyAddress)
            return HttpUtils.httpConnection(
                    mContext, token,
                    mmscUrl,
                    pdu, HttpUtils.HTTP_POST_METHOD,
                    mTransactionSettings.isProxySet,
                    mTransactionSettings.proxyAddress,
                    mTransactionSettings.proxyPort)
        }
    }

    private class NotifyRespTask internal constructor(context: Context, ind: NotificationInd, settings: TransactionSettings) : CommonAsyncTask(context, settings, ind) {

        private val context = WeakReference<Context>(context)

        override fun doInBackground(vararg params: Void): Void? {
            // Create the M-NotifyResp.ind
            var notifyRespInd: NotifyRespInd? = null
            try {
                notifyRespInd = NotifyRespInd(
                        PduHeaders.CURRENT_MMS_VERSION,
                        mNotificationInd.transactionId,
                        STATUS_RETRIEVED)

                val mContext = this.context.get()

                // Pack M-NotifyResp.ind and send it
                if (com.android.mms.MmsConfig.getNotifyWapMMSC()) {
                    sendPdu(PduComposer(mContext, notifyRespInd).make(), mContentLocation)
                } else {
                    sendPdu(PduComposer(mContext, notifyRespInd).make())
                }
            } catch (e: MmsException) {
                Log.e(TAG, "error", e)
            } catch (e: IOException) {
                Log.e(TAG, "error", e)
            }

            return null
        }
    }

    private class AcknowledgeIndTask internal constructor(context: Context, ind: NotificationInd, settings: TransactionSettings, private val mRetrieveConf: RetrieveConf) : CommonAsyncTask(context, settings, ind) {

        private val context = WeakReference<Context>(context)

        override fun doInBackground(vararg params: Void): Void? {
            // Send M-Acknowledge.ind to MMSC if required.
            // If the Transaction-ID isn't set in the M-Retrieve.conf, it means
            // the MMS proxy-relay doesn't require an ACK.
            val tranId = mRetrieveConf.transactionId
            if (tranId != null) {
                Log.v(TAG, "sending ACK to MMSC: " + mTransactionSettings.mmscUrl)
                // Create M-Acknowledge.ind
                var acknowledgeInd: com.google.android.mms.pdu_alt.AcknowledgeInd? = null

                try {
                    acknowledgeInd = com.google.android.mms.pdu_alt.AcknowledgeInd(
                            PduHeaders.CURRENT_MMS_VERSION, tranId)

                    val mContext = context.get()

                    // insert the 'from' address per spec
                    val lineNumber = Utils.getMyPhoneNumber(mContext)
                    acknowledgeInd.from = EncodedStringValue(lineNumber)

                    // Pack M-Acknowledge.ind and send it
                    if (com.android.mms.MmsConfig.getNotifyWapMMSC()) {
                        sendPdu(PduComposer(mContext, acknowledgeInd).make(), mContentLocation)
                    } else {
                        sendPdu(PduComposer(mContext, acknowledgeInd).make())
                    }
                } catch (e: InvalidHeaderValueException) {
                    Log.e(TAG, "error", e)
                } catch (e: MmsException) {
                    Log.e(TAG, "error", e)
                } catch (e: IOException) {
                    Log.e(TAG, "error", e)
                }

            }
            return null
        }
    }

    private fun getNotificationTask(context: Context, intent: Intent, response: ByteArray): List<CommonAsyncTask>? {
        if (response.size == 0) {
            Log.v(TAG, "MmsReceivedReceiver.sendNotification blank response")
            return null
        }

        if (mmscInfoForReceptionAck == null) {
            Log.v(TAG, "No MMSC information set, so no notification tasks will be able to complete")
            return null
        }

        val pdu = PduParser(response, MmsConfig.Overridden(MmsConfig(context), null).supportMmsContentDisposition).parse()
        if (pdu == null || pdu !is RetrieveConf) {
            android.util.Log.e(TAG, "MmsReceivedReceiver.sendNotification failed to parse pdu")
            return null
        }

        try {
            val ind = getNotificationInd(context, intent)
            val mmsc = mmscInfoForReceptionAck
            val transactionSettings = TransactionSettings(mmsc!!.mmscUrl, mmsc.mmsProxy, mmsc.proxyPort)

            val responseTasks = ArrayList<CommonAsyncTask>()
            responseTasks.add(AcknowledgeIndTask(context, ind, transactionSettings, pdu))
            responseTasks.add(NotifyRespTask(context, ind, transactionSettings))

            return responseTasks
        } catch (e: MmsException) {
            Log.e(TAG, "error", e)
            return null
        }

    }

    class MmscInformation(internal var mmscUrl: String, internal var mmsProxy: String, internal var proxyPort: Int)

    companion object {
        private val TAG = "MmsReceivedReceiver"

        val MMS_RECEIVED = "com.klinker.android.messaging.MMS_RECEIVED"
        val EXTRA_FILE_PATH = "file_path"
        val EXTRA_LOCATION_URL = "location_url"
        val EXTRA_TRIGGER_PUSH = "trigger_push"
        val EXTRA_URI = "notification_ind_uri"

        private val LOCATION_SELECTION =
                Telephony.Mms.MESSAGE_TYPE + "=? AND " + Telephony.Mms.CONTENT_LOCATION + " =?"

        private val RECEIVE_NOTIFICATION_EXECUTOR = Executors.newSingleThreadExecutor()

        @Throws(MmsException::class)
        private fun getNotificationInd(context: Context, intent: Intent): NotificationInd {
            return PduPersister.getPduPersister(context).load(intent.getParcelableExtra<Parcelable>(EXTRA_URI) as Uri) as NotificationInd
        }
    }
}
