package xyz.johnny.norntalk.messages

import android.content.Context
import android.graphics.Bitmap
import android.graphics.BitmapFactory
import android.net.Uri
import com.klinker.android.send_message.Message
import xyz.johnny.norntalk.database.NornDatabase
import xyz.johnny.norntalk.database.entities.Media
import xyz.johnny.norntalk.security.Curve
import java.io.ByteArrayOutputStream
import java.io.File
import java.io.FileInputStream
import java.io.FileOutputStream


/**
 * Classe permettant de gérer un document contenu dans un MMS
 */
class NornMedia(context: Context, val mimeType: String, val uri: String) {

    private var _data: ByteArray? = null
    val bitmap by lazy { this.getBitmap(context) }
    private var _name: String = "media_" + System.currentTimeMillis()
    val name get() = _name

    constructor(context: Context, mimeType: String, uri: Uri) : this(context, mimeType, uri.toString())


    private val data: ByteArray by lazy {
        this._data ?: Message.bitmapToByteArray(this.bitmap)
    }

    private fun readFileFromUri(context: Context, uri: Uri): ByteArray {
        val data = ByteArrayOutputStream()
        val tempByteArray = ByteArray(1024*1024)
        context.contentResolver.openFileDescriptor(uri, "r").use { desc ->
            FileInputStream(desc.fileDescriptor).use { inputStream ->
                inputStream.read(tempByteArray)
                data.write(tempByteArray)
            }
        }
        return data.toByteArray()
    }

    private fun getBitmap(context: Context): Bitmap {
        if (this.uri.startsWith("content:")) {
            context.contentResolver.openFileDescriptor(Uri.parse(this.uri), "r").use { parcelFileDescriptor ->
                val fileDescriptor = parcelFileDescriptor.fileDescriptor
                return BitmapFactory.decodeFileDescriptor(fileDescriptor)
            }
        } else {
            FileOutputStream(File(context.filesDir, this.uri)).use { fileOutputStream ->
                return BitmapFactory.decodeFileDescriptor(fileOutputStream.fd)
            }
        }
    }

    companion object {

        fun from(context: Context, media: Media): NornMedia {
            return NornMedia(context, media.type, media.uri)
        }

    }

    /**
     * Insérer le document dans la base de données
     *
     * @param context Contexte courant
     * @param message Message auquel le [Media] est associé
     * @param position position de ce [Media] pour le message
     */
    fun insert(context: Context, message: NornMessage, position: Int) {
        NornDatabase.Helper(context).run { db ->
            val media = Media(this.mimeType, this.uri, message.id, this.name, position)
            db.mediaDao().insertMedia(media)
        }
    }

    /**
     * Chiffre le contenu du message
     *
     * @param context Contexte courant
     * @param contact Destinataire du message
     * @return document chiffré
     */
    fun encrypt(context: Context, contact: NornContact): ByteArray {
        val curve = Curve.getCurve(context)
        val cipher = curve.getCipher(contact.pubKey!!)
        return cipher.encrypt(this.data)
    }

    /**
     * Chiffre le contenu du message en fonction de la conversation
     *
     * @param context Contexte courant
     * @param message Message auquel le document est associé
     */
    fun getContent(context: Context, message: NornMessage): ByteArray? {
        return if (message.conversation.secured && message.ciphertext != null)
            this.encrypt(context, message.contact)
        else
            this.data
    }

    /**
     * Déchiffre le document
     *
     * @param context Contexte courant
     * @param pubKey clé publique utilisée lors du chiffrement du message
     * @return document déchiffré
     */
    fun decrypt(context: Context, pubKey: ByteArray): ByteArray {
        val curve = Curve.getCurve(context)
        val cipher = curve.getCipher(pubKey)
        return cipher.decrypt(this.data)
    }

    override fun toString(): String {
        return "NornMedia(mimeType='$mimeType', uri=$uri, name='$_name')"
    }

}