package xyz.johnny.norntalk.messages

import android.content.Context
import android.os.AsyncTask
import android.telephony.SmsManager
import android.util.Base64
import android.util.Log
import org.bouncycastle.util.encoders.Hex
import xyz.johnny.norntalk.ConversationActivity
import xyz.johnny.norntalk.R
import xyz.johnny.norntalk.database.NornDatabase
import xyz.johnny.norntalk.database.entities.Message
import xyz.johnny.norntalk.security.Curve
import java.util.*
import kotlin.collections.ArrayList


/**
 * Classe liée aux message [Message].
 * Permet d'effectuer les principales opérations liées aux conversations
 *
 * @param text Contenu du message
 * @property ciphertext Message chiffré
 * @param sender expéditeur du message, null si envoyé par l'utilisateur
 * @property conversation Conversation du message
 * @property date date d'envoi du message si le message provient de l'utilisateur, date de reception sinon
 * @param context Contexte courant
 * @constructor Crée un message
 */
class NornMessage constructor(text: String?, var ciphertext: String?, sender: String?,
                              val conversation: NornConversation, val date: Date, context: Context) {

    companion object {

        /**
         * Déchiffre le message
         *
         * @param ciphertext Texte chiffré
         * @param context Contexte courant
         * @param key clé publique
         * @return Le message déchiffré
         */
        fun decrypt(ciphertext: String, context: Context, key: ByteArray): String {
            val curve = Curve.getCurve(context)
            val cipher = curve.getCipher(key)

            // si le déchiffrement échoue on assume que le contenu n'est pas chiffré
            return try {
                cipher.decrypt(Base64.decode(ciphertext, Base64.DEFAULT)).toString(Charsets.UTF_8)
            } catch (e: IllegalArgumentException) {
                ciphertext
            }
        }

        /**
         * Crée le message à partir d'un [Message] provenant de la base de données
         *
         * @param dbMessage [Message] provenant de la base de données
         * @param context Contexte courant
         * @return Le message
         */
        fun fromDbMessage(dbMessage: Message, context: Context): NornMessage {

            val key = dbMessage.key
            val text: String
            val ciphertext: String?
            if (key == null) {
                text = dbMessage.text
                ciphertext = null
            } else {
                ciphertext = dbMessage.text
                text = decrypt(ciphertext, context, key)
            }
            val message = NornMessage(text, ciphertext, dbMessage.contactNumber,
                    NornConversation.getConversation(dbMessage.conversationId, context),
                    dbMessage.date, context)

            message._dbMessage = dbMessage
            message.conversation.messages[dbMessage.id] = message
            return message
        }

        val smsManager by lazy { SmsManager.getDefault() }

    }

    private var _text = text?.trimEnd()
    /**
     * Contenu lisible du message, si le message est chiffré cet attribut sera déchiffré
     */
    val text get() = _text!!

    private var _rawtext: String? = null
    /**
     * Contenu d'origine du message
     */
    val rawtext get() = _rawtext!!

    /**
     * Expéditeur du message, null si le message a été envoyé par l'utilisateur
     */
    val sender: NornContact?

    /**
     * Contact associé au message.
     * Le destinataire si le message a été envoyé par l'utilisateur, sinon l'expéditeur
     */
    val contact: NornContact

    /**
     * Clé publique associée au message
     */
    var key: ByteArray? = null

    private var _dbMessage: Message? = null
    /**
     * [Message] dans la base de données
     */
    val dbMessage get() = _dbMessage!!
    /**
     * Vrai si le message a été lu, Faux sinon
     */
    val read get() = dbMessage.read
    /**
     * Identifiant du message
     */
    val id get() = dbMessage.id

    /**
     * Array contenant le numéro de téléphone de chaque destinataire
     */
    val addresses get() = conversation.members.map { it.number }.toTypedArray()

    private var _medias : ArrayList<NornMedia>? = null
    private val mediaLock = Object()
    /**
     * Medias attachés au message
     */
    val medias get(): ArrayList<NornMedia> {
        synchronized(mediaLock) {
            return _medias!!
        }
    }

    /**
     * Sujet du MMS
     */
    var subject: String = ""

    init {
        if (sender == null) {
            this.sender = null
            this.contact = this.conversation.members[0]
        } else {
            this.sender = NornContact.getContact(sender, context)
            this.contact = this.sender
        }
        if (this._dbMessage == null) {
            this._medias = ArrayList()
        } else {
            NornDatabase.Helper(context).run { db ->
                synchronized(mediaLock) {
                    val medias = db.mediaDao().getMediasFromMessage(this.id)
                    this._medias = ArrayList(medias.map { NornMedia.from(context, it) })
                }
            }
        }
    }

    /**
     * Créer un message envoyé par l'utilisateur
     *
     * @param text Contenu du message
     * @param conversation Conversation dans lequel envoyer le message
     * @param date Date de création du message
     * @param context Contexte courant
     */
    constructor(text: String, conversation: NornConversation, date: Date, context: Context) :
            this(text, null, null, conversation, date, context) {
        this._rawtext = text
        if (this.conversation.secured && this.contact.pubKey != null) {
            this.ciphertext = this.encrypt(context)
        }
    }

    /**
     * Créer un message reçu par l'utilisateur
     *
     * @param ciphertext Contenu du message
     * @param sender Numéro de téléphone de l'expéditeur
     * @param conversation Conversation associée au message
     * @param date Date de reception du message
     * @param context Contexte courant
     */
    constructor(ciphertext: String, sender: String, conversation: NornConversation, date: Date, context: Context) :
            this(null, ciphertext, sender, conversation, date, context) {
        this._rawtext = ciphertext

        if (this.conversation.secured && this.contact.pubKey != null) {
            this._text = decrypt(ciphertext, context, this.contact.pubKey!!)
        } else {
            this._text = ciphertext
            this.ciphertext = null
        }
    }

    /**
     * Insére le message dans la base de données
     *
     * @param context Contexte courant
     * @return [Message] dans la base de données
     */
    fun insertMessage(context: Context): AsyncTask<Void, Void, Message> {
        // ajouter la clé du message si la conversation est sécurisée
        this.key = if (this.conversation.secured) this.contact.pubKey else null

        // considéré le message comme lu si l'écran de la conversation est ouvert
        val read = this.conversation.id in ConversationActivity.foreground_activities

        val ciphertext = this.ciphertext
        val text = if (ciphertext == null) this.text else ciphertext

        // insérer le message dans la base de données
        return NornDatabase.Helper(context).run { db ->
            val message = Message(
                text,
                this.date,
                read,
                this.conversation.id,
                this.key,
                this.sender?.number
            )

            val rowid = db.messageDao().insertMessage(message)
            Log.d(this::class.java.simpleName, message.toString())

            // Ajouter l'entrée de la base de données à l'instance
            this._dbMessage = db.messageDao().getMessageByRowId(rowid)

            // Ajouter le message à la conversation
            this.conversation.messages[this.id] = this

            if (this.medias.isNotEmpty()) {
                for (i in this.medias.indices) {
                    this.medias[i].insert(context, this, i)
                }
            }

            // retourner le message
            this.dbMessage
        }
    }

    /**
     * Supprimer le message
     *
     * @param context Contexte courant
     */
    fun delete(context: Context) {
        NornDatabase.Helper(context).run { db ->
            db.messageDao().deleteMessage(this.dbMessage)
        }
        NornMessageDispatcher.delete(arrayOf(this))
    }

    /**
     * Envoyer une notification pour le message courant
     */
    fun sendNotification() = this.conversation.notification.send(this)

    override fun toString(): String {
        return """Message[id=${this._dbMessage?.id}, sender=${this.sender},
            | text=${this.text.trim()}, cipĥertext=${this.ciphertext?.trim()},
            | read=${this._dbMessage?.read}, contact=${this.contact.number}, date=${this.date},
            | conversation=${this.conversation}]
            | medias=${this._medias.toString()}""".trimMargin()
    }

    /**
     * Envoie le message
     *
     * @param context Contexte courant
     * @param insert Si Vrai le message doit être inséré dans la base de données, Faux sinon
     *               cela permet de cacher les messages d'échange de clés à l'utilisateur
     * @param raw Si Vrai le message sera envoyé en clair, sinon le message sera chiffré en fonction
     *            des propriétés de la conversation
     */
    fun send(context: Context, insert: Boolean = true, raw: Boolean = false): Boolean {
        Log.d(this::class.java.simpleName, "sending: " + this)
        val transaction = NornTransaction(context)
        return transaction.sendNewMessage(this, insert, raw)
    }

    /**
     * Chiffre le contenu du message
     *
     * @param context Contexte courant
     * @return message chiffré
     */
    fun encrypt(context: Context): String {
        val curve = Curve.getCurve(context)
        val cipher = curve.getCipher(contact.pubKey!!)
        return Base64.encodeToString(cipher.encrypt(this.text.toByteArray(Charsets.UTF_8)), Base64.DEFAULT)
    }

    /**
     * Lit la clé publique provenant d'un message commençant par "ke:" et l'associe au contact
     * concerné par ce message
     *
     * @param context Contexte courant
     */
    fun setPublicKey(context: Context) {
        val pubkey = this.rawtext.substring(context.getString(R.string.key_exchange_prefix).length)
        Log.d(this::class.java.simpleName, pubkey + ": " + pubkey.length)
        this.sender!!.setPublicKey(Hex.decode(pubkey), context)
    }

    /**
     * Envoyer le message chiffré en fonction de la conversation
     *
     * @param raw forcer l'envoi du message en clair
     */
    fun messageBody(raw: Boolean) =
            if (this.conversation.secured && this.ciphertext != null && !raw) this.ciphertext!!
            else this.text

    override fun equals(other: Any?): Boolean {
        if (this === other) return true
        if (javaClass != other?.javaClass) return false

        other as NornMessage
        if (this._dbMessage == null || other._dbMessage == null) return false
        return this.dbMessage.id != other.dbMessage.id
    }

    override fun hashCode(): Int {
        return _dbMessage?.hashCode() ?: 0
    }

}