package xyz.johnny.norntalk.messages

import android.util.SparseArray
import xyz.johnny.norntalk.ui_components.BaseNornAdapter
import xyz.johnny.norntalk.ui_components.NornListAdapter

/**
 * Cet objet permet de distribuer les messages aux [BaseNornAdapter] nécessitant d'obtenir les
 * messages en temps réel
 */
object NornMessageDispatcher {

    /**
     * Association Identifiant de conversation/Ensemble de [BaseNornAdapter]
     */
    private val subscribers = SparseArray<HashSet<BaseNornAdapter>>()
    /**
     * Retourne les [NornListAdapter], correspondant à la liste des conversations
     */
    val mainAdapter get() = subscribers[NornListAdapter.conversationId]

    /**
     * Ajoute un [BaseNornAdapter] dans l'association [subscribers]
     *
     * @param adapter [BaseNornAdapter] à insérer
     */
    @Synchronized
    fun subscribe(adapter: BaseNornAdapter) {
        if (this.subscribers[adapter.conversationId] == null)
            this.subscribers.put(adapter.conversationId, HashSet())

        this.subscribers.get(adapter.conversationId)!!.add(adapter)
    }

    /**
     * Supprime un [BaseNornAdapter] de l'association [subscribers]
     *
     * @param adapter [BaseNornAdapter] à supprimer
     */
    @Synchronized
    fun unsubscribe(adapter: BaseNornAdapter) {
        this.subscribers[adapter.conversationId]?.remove(adapter)
    }

    /**
     * Transmet un message à l'ensemble des [BaseNornAdapter] concernés
     *
     * @param message Message à transmettre
     */
    fun send(message: NornMessage) {
        // transmettre aux NornListAdapters
        this.send(message, NornListAdapter.conversationId)
        // transmettre aux NornMessageAdapters concernés
        this.send(message, message.conversation.id)
    }

    /**
     * Transmet un message à l'ensemble des [BaseNornAdapter] pour une conversation donnée
     *
     * @param message Message à transmettre
     * @param convId Identifiant de la conversation
     */
    fun send(message: NornMessage, convId: Int) {
        val adapters = this.subscribers[convId] ?: return  // retourne si l'élément n'existe pas

        for (adapter in adapters)
            adapter.addMessage(message)
    }

    /**
     * Transmettre la suppression de messages aux [BaseNornAdapter] concernés
     *
     * @param messages Liste de message supprimés
     */
    fun delete(messages: Array<NornMessage>) {
        this.delete(messages, NornListAdapter.conversationId)

        // trier les messages en fonction de leur conversation
        val conversations = HashMap<Int, ArrayList<NornMessage>>()

        for (message in messages) {
            if (message.conversation.id !in conversations) {
                conversations.put(message.conversation.id, ArrayList())
            }
            conversations.get(message.conversation.id)!!.add(message)
        }

        // supprimer les messages dans les conversations qui les concerne
        conversations.forEach { conversation: Map.Entry<Int, ArrayList<NornMessage>> ->
            this.delete(conversation.value.toTypedArray(), conversation.key)
        }
    }

    /**
     * Supprimer les messages pour une conversation donnée
     *
     * @param messages Message à supprimer
     * @param convId Identifiant de la conversation
     */
    fun delete(messages: Array<NornMessage>, convId: Int) {
        val adapters = this.subscribers[convId] ?: return

        for (adapter in adapters) {
            adapter.deleteMessages(messages)
        }
    }

    /**
     * Supprimer une conversation entière
     *
     * @param convId Identifiant de la conversation à supprimer
     */
    fun delete(convId: Int) {
        val adapters = this.subscribers[NornListAdapter.conversationId] ?: return

        for (adapter in adapters) {
            if (adapter is NornListAdapter)
                adapter.delete(convId)
        }
    }

}