package xyz.johnny.norntalk.messages

import android.annotation.SuppressLint
import android.annotation.TargetApi
import android.app.NotificationChannel
import android.app.NotificationManager
import android.app.PendingIntent
import android.content.Context
import android.content.Intent
import android.os.Build
import android.preference.PreferenceManager
import android.support.annotation.RequiresApi
import android.support.v4.app.NotificationCompat
import android.support.v4.app.RemoteInput
import android.util.Log
import android.util.SparseArray
import android.widget.Toast
import xyz.johnny.norntalk.ConversationActivity
import xyz.johnny.norntalk.R
import java.util.*


/**
 * Classe représentant une notification.
 * Il existe une notification par conversation.
 *
 * @param conversationId Identifiant de la conversation concernée par la notification
 * @property context Contexte courant
 * @constructor Crée une notification pour la conversation courante en prenant en compte les
 *              paramètres de l'utilisateur
 */
class NornNotification private constructor(val conversationId: Int, val context: Context) {

    companion object {

        /**
         * Association Identifiant de conversation/[NornNotification]
         */
        private val notifications = SparseArray<NornNotification>()

        /**
         * Le pattern des vibrations des notifications
         */
        private val vibrationPattern = longArrayOf(300, 300, 300, 300, 300)

        private var _channelMessages: NotificationChannel? = null
        /**
         * Instance de NotificationChannel pour Android O
         */
        private val channelMessages get() = _channelMessages!!

        private const val channelId = "messages"

        @RequiresApi(Build.VERSION_CODES.O)
        fun updateChannel(context: Context) {
            if (this._channelMessages == null) {
                createChannel(context)
            } else {
                channelMessages.name = context.getString(R.string.channel_messages)
                channelMessages.description = context.getString(R.string.channel_description_messages)

                // register the notification channel
                val notificationManager = context.getSystemService(NotificationManager::class.java)
                notificationManager.createNotificationChannel(channelMessages)
            }
        }

        @RequiresApi(Build.VERSION_CODES.O)
        /**
         * Crée la chaîne de notification sur Android O
         *
         * @param context Contexte courant
         */
        fun createChannel(context: Context) {
            val name = context.getString(R.string.channel_messages)
            val importance = NotificationManager.IMPORTANCE_HIGH
            _channelMessages = NotificationChannel(channelId, name, importance)

            channelMessages.enableLights(true)
            channelMessages.lightColor = context.getColor(R.color.colorPrimary)

            channelMessages.enableVibration(true)
            channelMessages.vibrationPattern = vibrationPattern

            this.updateChannel(context)
        }


        @TargetApi(Build.VERSION_CODES.O)
        /**
         * Méthode retournant un builder pour différentes versions d'Android
         *
         * @param context Contexte courant
         */
        fun getBuilder(context: Context): NotificationCompat.Builder {
            Log.d(this::class.java.simpleName, "build: %d %d".format(Build.VERSION.SDK_INT, Build.VERSION_CODES.O))
            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O)
                createChannel(context)

            val builder = NotificationCompat.Builder(context, channelId)
            builder.setLights(context.getColor(R.color.colorPrimary), 2000, 500)

            if (Build.VERSION.SDK_INT < Build.VERSION_CODES.O) {
                builder.setVibrate(vibrationPattern)
                builder.priority = NotificationCompat.PRIORITY_HIGH
            }

            // assigner une icone, un pattern de vibrations et de clignottement de la LED du téléphone
            builder.setSmallIcon(R.drawable.ic_message)

            return builder
        }

        /**
         * Retourne la notification associée à une conversation
         *
         * @param conversation La conversation
         * @param context Contexte courant
         * @return [NornNotification] correspondante
         */
        @Synchronized
        fun getNotification(conversation: NornConversation, context: Context): NornNotification {
            val notification = notifications[conversation.id]
            if (notification != null)
                return notification

            return NornNotification(conversation.id, context.applicationContext)
        }

        /**
         * Notification toast, accessible à travers [getToast]
         */
        private var toast: Toast? = null

        /**
         * Crée une notification "toast" pouvant être modifié par la suite pour notifier de l'envoi
         * et de la reception des messages
         *
         * @param context Contexte courant
         * @return Notification toast
         */
        @SuppressLint("ShowToast")
        @Synchronized
        fun getToast(context: Context): Toast {
            if (this.toast == null)
                toast = Toast.makeText(context, "", Toast.LENGTH_SHORT)

            return this.toast!!
        }

    }

    /**
     * Conversation concernée par la notification
     */
    val conversation get() = NornConversation.getConversation(conversationId, context)

    /**
     * L'instance du builder de la notification
     */
    val builder = getBuilder(context)
    /**
     * Les messages affichés dans la notification
     */
    val messages = ArrayList<NornMessage>()
    /**
     * Style utilisé pour afficher les messages
     */
    private val messagingStyle = NotificationCompat.MessagingStyle(context.getString(R.string.me))
    /**
     * Le gestionnaire de notifications
     */
    val notificationManager = context.getSystemService(Context.NOTIFICATION_SERVICE) as NotificationManager
    /**
     * Préférences de l'application, pour s'adapter aux options choisies par l'utilisateur
     */
    val preferences = PreferenceManager.getDefaultSharedPreferences(context)


    init {
        // conserver la notification dans l'association
        notifications.put(conversationId, this)

        // assigner le style à la notification
        this.builder.setStyle(this.messagingStyle)

        // Ouvrir l'application lorsqu'on appuie sur la notification
        val intent = Intent(context, ConversationActivity::class.java)
        intent.putExtra("conversationId", this.conversation.id)

        val piConv = PendingIntent.getActivity(context.applicationContext, this.conversation.id,
                                               intent, PendingIntent.FLAG_CANCEL_CURRENT)
        this.builder.setContentIntent(piConv)

        // créer l'action de réponse rapide
        val KEY_TEXT_REPLY = "text_reply"
        val replyLabel = context.getString(R.string.reply_label)
        val remoteInput = RemoteInput.Builder(KEY_TEXT_REPLY).setLabel(replyLabel).build()

        val messageReplyIntent = Intent(context, NornMessageService::class.java)
        messageReplyIntent.putExtra("conversationId", conversation.id)
        messageReplyIntent.action = NornMessageService.REPLY_ACTION

        val replyPendingIntent = PendingIntent.getService(context.applicationContext,
                conversation.id, messageReplyIntent, PendingIntent.FLAG_UPDATE_CURRENT)

        val actionReply = NotificationCompat.Action.Builder(R.drawable.ic_reply_icon,
                                                            context.getString(R.string.reply_label), replyPendingIntent)
        actionReply.addRemoteInput(remoteInput)

        // créer l'action marquer comme lu
        val readIntent = Intent(context, NornMessageService::class.java)
        readIntent.putExtra("conversationId", this.conversation.id)
        readIntent.action = NornMessageService.READ_ACTION

        val readPendingIntent = PendingIntent.getService(context.applicationContext,
                conversation.id, readIntent, PendingIntent.FLAG_UPDATE_CURRENT)

        val actionRead = NotificationCompat.Action.Builder(R.drawable.ic_clear,
                                                           context.getString(R.string.read), readPendingIntent)

        // Ajouter les actions à la notification
        this.builder.addAction(actionReply.build())
        this.builder.addAction(actionRead.build())
    }

    /**
     * Supprimer la notification et les messages qu'elle contient
     */
    fun clear() {
        notificationManager.cancel(this.conversation.id)
        this.messagingStyle.messages.clear()
    }

    /**
     * Afficher la notification à l'utilisateur ou ajouter un nouveau message à une notification
     * déjà existante
     *
     * @param msg Nouveau message
     */
    fun send(msg: NornMessage) {
        // ne pas afficher la notification si l'utilisateur a décidé d'ignorer la conversation
        if (this.conversation.muted)
            return

        // ne pas afficher la notification si l'utilisateur est déjà sur l'écran de la conversation
        if (this.conversation.id in ConversationActivity.foreground_activities)
            return

        this.messages.add(msg)

        // cacher le contenu du message si l'utilisateur le souhaite
        val text = if (this.preferences.getBoolean(context.getString(R.string.notifications_content), true))
            msg.text
        else
            context.getString(R.string.hidden_text)

        // cacher le nom de l'expéditeur si l'utilisateur le souhaite
        val sender = if (this.preferences.getBoolean(this.context.getString(R.string.notifications_contact), true))
            msg.sender!!.name
        else
            this.context.getString(R.string.hidden_conversation)

        // ajouter le message à la notification
        val msMsg = NotificationCompat.MessagingStyle.Message(text, msg.date.time, sender)
        this.messagingStyle.addMessage(msMsg)

        // afficher la notification
        this.notificationManager.notify(this.conversation.id, this.builder.build())

    }

}