/*
 * Copyright 2013 Jacob Klinker
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package xyz.johnny.norntalk.messages

import android.app.Activity
import android.app.PendingIntent
import android.content.ContentResolver
import android.content.Context
import android.content.Intent
import android.net.Uri
import android.os.Bundle
import android.os.Looper
import android.telephony.SmsManager
import android.telephony.SmsMessage
import android.text.TextUtils
import android.view.View
import android.widget.Toast
import com.android.mms.MmsConfig
import com.android.mms.dom.smil.parser.SmilXmlSerializer
import com.android.mms.util.DownloadManager
import com.android.mms.util.RateController
import com.google.android.mms.ContentType
import com.google.android.mms.InvalidHeaderValueException
import com.google.android.mms.MMSPart
import com.google.android.mms.pdu_alt.*
import com.google.android.mms.smil.SmilHelper
import com.klinker.android.logger.Log
import com.klinker.android.send_message.*
import java.io.ByteArrayOutputStream
import java.io.File
import java.io.FileOutputStream
import java.io.IOException
import java.util.*

/**
 * Class to process transaction requests for sending
 *
 * @author Jake Klinker
 *
 * Sets context and settings
 *
 * @param context  is the context of the activity or service
 */
class NornTransaction constructor(private val context: Context) {

    private var explicitSentSmsReceiver: Intent? = null
    private var explicitSentMmsReceiver: Intent? = null
    private var explicitDeliveredSmsReceiver: Intent? = null

    private var saveMessage = true

    var SMS_SENT = ".SMS_SENT"
    var SMS_DELIVERED = ".SMS_DELIVERED"

    init {
        SMS_SENT = context.packageName + SMS_SENT
        SMS_DELIVERED = context.packageName + SMS_DELIVERED

        if (NOTIFY_SMS_FAILURE == ".NOTIFY_SMS_FAILURE") {
            NOTIFY_SMS_FAILURE = context.packageName + NOTIFY_SMS_FAILURE
        }
    }

    /**
     * Evoyer un message à partir d'une instance de [NornMessage]
     *
     * @param message Message à envoyer
     * @param insert Si Vrai le message doit être inséré dans la base de données, Faux sinon
     *               cela permet de cacher les messages d'échange de clés à l'utilisateur
     * @param raw Si Vrai le message sera envoyé en clair, sinon le message sera chiffré en fonction
     *            des propriétés de la conversation
     */
    fun sendNewMessage(message: NornMessage, insert: Boolean, raw: Boolean): Boolean {
        if (message.medias.isEmpty()) {
            if (message.text.isBlank())
                return false
            else
                this.sendSmsMessage(message, insert, raw)
        } else {
            try {
                Looper.prepare()
            } catch (e: Exception) {}

            RateController.init(context)
            DownloadManager.init(context)
            this.sendMmsMessage(message)
        }
        return true
    }

    private fun sendSmsMessage(message: NornMessage, insert: Boolean, raw: Boolean) {
        // envoyer le message chiffré si la conversation est sécurisée
        val body = message.messageBody(raw)

        // save the message for each of the addresses
        for (i in message.addresses.indices) {
            var sentPI : PendingIntent? = null
            var deliveredPI : PendingIntent? = null

            if (insert) {
                message.insertMessage(context).get()
                Log.v("send_transaction", "message id: " + message.id)

                // set up sent and delivered pending intents to be used with message request
                val sentIntent: Intent?
                if (explicitSentSmsReceiver == null) {
                    sentIntent = Intent(SMS_SENT)
                    BroadcastUtils.addClassName(context, sentIntent, SMS_SENT)
                } else {
                    sentIntent = explicitSentSmsReceiver
                }

                sentIntent!!.putExtra("message_id", message.id)
                sentPI = PendingIntent.getBroadcast(context, message.id.toInt(), sentIntent,
                                                    PendingIntent.FLAG_UPDATE_CURRENT)

                val deliveredIntent: Intent?
                if (explicitDeliveredSmsReceiver == null) {
                    deliveredIntent = Intent(SMS_DELIVERED)
                    BroadcastUtils.addClassName(context, deliveredIntent, SMS_DELIVERED)
                } else {
                    deliveredIntent = explicitDeliveredSmsReceiver
                }

                deliveredIntent!!.putExtra("message_id", message.id)
                deliveredPI = PendingIntent.getBroadcast(context, message.id.toInt(),
                        deliveredIntent, PendingIntent.FLAG_UPDATE_CURRENT)
            }

            val sPI = ArrayList<PendingIntent?>()
            val dPI = ArrayList<PendingIntent?>()

            val smsManager = SmsManagerFactory.createSmsManager(settings)
            Log.v("send_transaction", "found sms manager")

            if (settings.split) {
                Log.v("send_transaction", "splitting message")
                // figure out the length of supported message
                val splitData = SmsMessage.calculateLength(body, false)

                // we take the current length + the remaining length to get the total number of characters
                // that message set can support, and then divide by the number of message that will require
                // to get the length supported by a single message
                var length = (body.length + splitData[2]) / splitData[0]
                Log.v("send_transaction", "length: $length")

                var counter = false
                if (settings.splitCounter && body.length > length) {
                    counter = true
                    length -= 6
                }

                // get the split messages
                val textToSend = splitByLength(body, length, counter)

                // send each message part to each recipient attached to message
                for (j in textToSend.indices) {
                    val parts = smsManager.divideMessage(textToSend[j])

                    for (k in parts.indices) {
                        sPI.add(sentPI)
                        dPI.add(if (settings.deliveryReports) deliveredPI else null)
                    }

                    Log.v("send_transaction", "sending split message")
                    smsManager.sendMultipartTextMessage(message.addresses[i], null, parts, sPI, dPI)
                }
            } else {
                Log.v("send_transaction", "sending without splitting")
                // send the message normally without forcing anything to be split
                val parts = smsManager.divideMessage(body)

                for (j in parts.indices) {
                    sPI.add(sentPI)
                    dPI.add(if (settings.deliveryReports) deliveredPI else null)
                }

                try {
                    Log.v("send_transaction", "sent message")
                    smsManager.sendMultipartTextMessage(message.addresses[i], null, parts, sPI, dPI)
                } catch (e: Exception) {
                    // whoops...
                    Log.v("send_transaction", "error sending message")
                    Log.e(TAG, "exception thrown", e)

                    try {
                        (context as Activity).window.decorView.findViewById<View>(android.R.id.content).post({ Toast.makeText(context, "Message could not be sent", Toast.LENGTH_LONG).show() })
                    } catch (f: Exception) {}

                }

            }
        }
    }

    private fun sendMmsMessage(message: NornMessage) {
        // create the medias to send
        val data = ArrayList<MMSPart>()

        // add any extra medias according to their mimeType set in the message
        //      eg. videos, audio, contact cards, location maybe?
        for (m in message.medias) {
            val part = MMSPart()
            part.Name = m.name
            part.MimeType = m.mimeType
            part.Data = m.getContent(context, message)
            data.add(part)
        }

        val text = message.messageBody(false)
        if (message.text != "") {
            // add text to the end of the part and send
            val part = MMSPart()
            part.Name = "text"
            part.MimeType = "text/plain"
            part.Data = text.toByteArray()
            data.add(part)
        }

        message.insertMessage(context).get()

        Log.v(TAG, "using system method for sending")
        sendMmsThroughSystem(context, message.subject, data, message.addresses, explicitSentMmsReceiver)
    }

    // splits text and adds split counter when applicable
    private fun splitByLength(s: String, chunkSize: Int, counter: Boolean): Array<String?> {
        val arraySize = Math.ceil(s.length.toDouble() / chunkSize).toInt()

        val returnArray = arrayOfNulls<String>(arraySize)

        var index = 0
        run {
            var i = 0
            while (i < s.length) {
                if (s.length - i < chunkSize) {
                    returnArray[index++] = s.substring(i)
                } else {
                    returnArray[index++] = s.substring(i, i + chunkSize)
                }
                i = i + chunkSize
            }
        }

        if (counter && returnArray.size > 1) {
            for (i in returnArray.indices) {
                returnArray[i] = "(" + (i + 1) + "/" + returnArray.size + ") " + returnArray[i]
            }
        }

        return returnArray
    }

    companion object {

        private val TAG = "NornTransaction"
        val settings by lazy { Settings() }

        var NOTIFY_SMS_FAILURE = ".NOTIFY_SMS_FAILURE"
        val MMS_ERROR = "com.klinker.android.send_message.MMS_ERROR"
        val REFRESH = "com.klinker.android.send_message.REFRESH"
        val MMS_PROGRESS = "com.klinker.android.send_message.MMS_PROGRESS"
        val NOTIFY_OF_DELIVERY = "com.klinker.android.send_message.NOTIFY_DELIVERY"
        val NOTIFY_OF_MMS = "com.klinker.android.messaging.NEW_MMS_DOWNLOADED"

        val NO_THREAD_ID: Long = 0

        val DEFAULT_EXPIRY_TIME = (7 * 24 * 60 * 60).toLong()
        val DEFAULT_PRIORITY = PduHeaders.PRIORITY_NORMAL

        private fun sendMmsThroughSystem(context: Context, subject: String, parts: List<MMSPart>,
                                         addresses: Array<String>, explicitSentMmsReceiver: Intent?) {
            val fileName = "send." + Math.abs(Random().nextLong()).toString() + ".dat"
            val mSendFile = File(context.cacheDir, fileName)

            val sendReq = buildPdu(context, addresses, subject, parts)

            val intent = Intent(context, NornMMSSentReceiver::class.java)

            intent.putExtra(MmsSentReceiver.EXTRA_FILE_PATH, mSendFile.path)
            val pendingIntent = PendingIntent.getBroadcast(
                    context, 0, intent, PendingIntent.FLAG_CANCEL_CURRENT)

            val writerUri = Uri.Builder()
                    .authority("xyz.johnny.norntalk.messages.NornMMSFileProvider")
                    .path(fileName)
                    .scheme(ContentResolver.SCHEME_CONTENT)
                    .build()
            var contentUri: Uri? = null

            val data = PduComposer(context, sendReq).make()

            try {
                FileOutputStream(mSendFile).use { writer ->
                    writer.write(data)
                }
                contentUri = writerUri
            } catch (e: IOException) {
                Log.e(TAG, "Error writing send file", e)
            }

            val configOverrides = Bundle()
            configOverrides.putBoolean(SmsManager.MMS_CONFIG_GROUP_MMS_ENABLED, settings.group)
            val httpParams = MmsConfig.getHttpParams()
            if (!TextUtils.isEmpty(httpParams)) {
                configOverrides.putString(SmsManager.MMS_CONFIG_HTTP_PARAMS, httpParams)
            }
            configOverrides.putInt(SmsManager.MMS_CONFIG_MAX_MESSAGE_SIZE, MmsConfig.getMaxMessageSize())

            if (contentUri != null) {
                SmsManager.getDefault().sendMultimediaMessage(context,
                        contentUri, null, configOverrides, pendingIntent)
            } else {
                Log.e(TAG, "Error writing sending Mms")
                pendingIntent.send(SmsManager.MMS_ERROR_IO_ERROR)
            }

        }

        private fun buildPdu(context: Context, recipients: Array<String>, subject: String,
                             parts: List<MMSPart>): SendReq {
            val req = SendReq()
            // From, per spec
            val lineNumber = Utils.getMyPhoneNumber(context)
            if (!TextUtils.isEmpty(lineNumber)) {
                req.from = EncodedStringValue(lineNumber)
            }
            // To
            for (recipient in recipients) {
                req.addTo(EncodedStringValue(recipient))
            }
            // Subject
            if (!TextUtils.isEmpty(subject)) {
                req.subject = EncodedStringValue(subject)
            }
            // Date
            req.date = System.currentTimeMillis() / 1000
            // Body
            val body = PduBody()
            // Add text part. Always add a smil part for compatibility, without it there
            // may be issues on some carriers/client apps
            var size = 0
            for (i in parts.indices) {
                val part = parts[i]
                size += addTextPart(body, part, i)
            }

            // add a SMIL document for compatibility
            val out = ByteArrayOutputStream()
            SmilXmlSerializer.serialize(SmilHelper.createSmilDocument(body), out)
            val smilPart = PduPart()
            smilPart.contentId = "smil".toByteArray()
            smilPart.contentLocation = "smil.xml".toByteArray()
            smilPart.contentType = ContentType.APP_SMIL.toByteArray()
            smilPart.data = out.toByteArray()
            body.addPart(0, smilPart)

            req.body = body
            // Message size
            req.messageSize = size.toLong()
            // Message class
            req.messageClass = PduHeaders.MESSAGE_CLASS_PERSONAL_STR.toByteArray()
            // Expiry
            req.expiry = DEFAULT_EXPIRY_TIME
            try {
                // Priority
                req.priority = DEFAULT_PRIORITY
                // Delivery report
                req.deliveryReport = PduHeaders.VALUE_NO
                // Read report
                req.readReport = PduHeaders.VALUE_NO
            } catch (e: InvalidHeaderValueException) {
            }

            return req
        }

        private fun addTextPart(pb: PduBody, p: MMSPart, id: Int): Int {
            val filename = p.Name
            val part = PduPart()
            // Set Charset if it's a text medias.
            if (p.MimeType.startsWith("text")) {
                part.charset = CharacterSets.UTF_8
            }
            // Set Content-Type.
            part.contentType = p.MimeType.toByteArray()
            // Set Content-Location.
            part.contentLocation = filename.toByteArray()
            val index = filename.lastIndexOf(".")
            val contentId = if (index == -1)
                filename
            else
                filename.substring(0, index)
            part.contentId = contentId.toByteArray()
            part.data = p.Data
            pb.addPart(part)

            return part.data.size
        }

    }

}
