package xyz.johnny.norntalk.security

import org.bouncycastle.crypto.engines.ChaChaEngine
import org.bouncycastle.crypto.params.KeyParameter
import org.bouncycastle.crypto.params.ParametersWithIV
import java.io.ByteArrayInputStream
import java.io.ByteArrayOutputStream

/**
 * Classe permettant de chiffrer et déchiffrer les messages de l'application selon l'algorithme ChaCha20.
 * Une clé de chiffrement doit être donnée.
 *
 * @property key Clé de chiffrement
 * @constructor Créé une instance de la classe ChaCha utilisant la clé de chiffrement donnée
 */
class ChaCha constructor(val key: ByteArray) {

    companion object {

        /**
         * Taille de la clé en octets
         */
        val KEY_SIZE = 32
        /**
         * Taille du vecteur d'initialisation en octets
         */
        private val IV_SIZE = 8
        /**
         * Taille des buffers utilisés pendant le chiffrement et le déchiffrement en octets
         */
        private val BUFFER_SIZE = 8192

        /**
         * Génère une clé de chiffrement pour l'algorithme ChaCha20
         *
         * @return Clé de chiffrement
         */
        fun generateKey(): ByteArray {
            val keyBytes = ByteArray(this.KEY_SIZE)
            Security.random.nextBytes(keyBytes)
            return keyBytes
        }

        /**
         * Génère un vecteur d'initialisation pour l'algorithme ChaCha20
         *
         * @return Vecteur d'initialisation
         */
        fun generateIV(): ByteArray {
            val ivBytes = ByteArray(this.IV_SIZE)
            Security.random.nextBytes(ivBytes)
            return ivBytes
        }

    }

    /**
     * La clé de chiffrement
     */
    private val keyParameter = KeyParameter(key)

    /**
     * Si le encrypt est Vrai alors le message text est chiffré
     * sinon le message text est déchiffré
     *
     * @param encrypt Vrai si le message doit être déchiffré, Faux sinon
     * @param text Message à chiffrer/déchiffrer
     * @return Le message chiffré ou déchiffré en fonction de la valeur de encrypt
     */
    private fun crypt(encrypt: Boolean, text: ByteArray): ByteArray {
        val input = ByteArrayInputStream(text)
        val output = ByteArrayOutputStream()

        val inputBuffer = ByteArray(BUFFER_SIZE)
        val outputBuffer = ByteArray(BUFFER_SIZE)

        // instanciation du vecteur d'initialisation
        val params: ParametersWithIV
        if (encrypt) {
            // générer le vecteur d'initialisation si le message doit être chiffré
            val iv = generateIV()
            output.write(iv)
            params = ParametersWithIV(keyParameter, iv)
        } else {
            // lire le vecteur d'initialisation lors du déchiffrement du message
            val iv = ByteArray(8)
            input.read(iv)
            params = ParametersWithIV(keyParameter, iv)
        }

        // chiffrement/déchiffrement du message
        val engine = ChaChaEngine()
        engine.init(encrypt, params)

        // lecture de la première partie du message
        var len = input.read(inputBuffer)
        while (len != -1) {
            // chiffrement/déchiffrement de la partie du message
            engine.processBytes(inputBuffer, 0, len, outputBuffer, 0)
            // écrire le texte chiffré/déchiffré dans le stream de sortie
            output.write(outputBuffer, 0, len)
            // lecture de la prochaine partie du message
            len = input.read(inputBuffer)
        }

        // retourner le message chiffré/déchiffré sous forme d'octets
        return output.toByteArray()

    }

    /**
     * Chiffre le message plaintext
     *
     * @param plaintext Message à chiffrer
     * @return Le message chiffré
     */
    fun encrypt(plaintext: ByteArray) = crypt(true, plaintext)

    /**
     * Déchiffre le message ciphertext
     *
     * @param ciphertext Message chiffré
     * @return Le message déchiffré
     */
    fun decrypt(ciphertext: ByteArray) = crypt(false, ciphertext)

}