package xyz.johnny.norntalk.security

import android.content.Context
import android.telephony.PhoneNumberUtils
import android.util.Log
import org.bouncycastle.util.encoders.Hex
import xyz.johnny.norntalk.messages.NornContact
import xyz.johnny.norntalk.messages.NornConversation
import xyz.johnny.norntalk.messages.NornMessage
import java.util.*

/**
 * Objet permettant de vérifier qu'un numéro de téléphone est bien celui de l'utilisateur
 */
object CheckNumber {

    var number : String = ""

    /**
     * Taille du token en octets
     */
    private val TOKEN_SIZE = 32

    /**
     * Token qui sera présent dans le message reçu
     */
    var token: String? = null
        private set(value) { field = value }

    private val numberObj = NumberObject()

    /**
     * Signale au thread executant checkNumber que le message envoyé a bien été reçu
     */
    fun messageReceived(msg: NornMessage): Boolean {
        if (this.token == msg.rawtext) {
            synchronized(this.numberObj) {
                Log.d(this::class.java.simpleName, "sender: " + msg.sender?.number)
                this.numberObj.number = msg.sender?.number ?: ""
                this.number = msg.sender?.number ?: ""
                this.numberObj.notifyAll()
            }
        }

        return this.token == msg.rawtext
    }

    /**
     * Vérifier qu'un numéro de téléphone est celui de l'utilisateur
     *
     * @param number Numéro de téléphone à vérifier
     * @param context Contexte courant
     * @return Vrai si le numéro est celui de l'utilisateur, Faux sinon
     */
    fun checkNumber(number: String?, context: Context): Boolean {
        Log.d(this::class.java.simpleName, number)
        // vérifier que le numéro est non null
        if (number == null)
            return false

        // vérifier que le numéro est valide
        if (!PhoneNumberUtils.isWellFormedSmsAddress(number))
            return false

        // utiliser le lock de numberObj
        synchronized(this.numberObj) {

            try {
                // générer le token
                this.token = this.genToken()

                // créer le message
                val contact = NornContact.getContact(number, context)
                val conversation = NornConversation.getConversation(arrayOf(contact), context)
                val message = NornMessage(this.token!!, conversation, Date(), context)

                // envoyer le message
                message.send(context, false, true)

                val end = System.currentTimeMillis() + 3000
                while (System.currentTimeMillis() < end)
                    this.numberObj.wait(end - System.currentTimeMillis())

            } finally {
                // réinitialiser le token et le lock
                this.token = null
            }

        }

        // Vrai si le numéro de téléphone du message reçu correspond au numéro en entrée de la méthode
        return this.numberObj.number == NornContact.format(number, context)
    }

    /**
     * Génère un token de longueur TOKEN_SIZE
     *
     * @return Token généré
     */
    private fun genToken(): String {
        val token = ByteArray(TOKEN_SIZE)
        Security.random.nextBytes(token)
        return Hex.toHexString(token)
    }

}