package xyz.johnny.norntalk.security

import android.content.Context
import android.util.Base64
import org.whispersystems.curve25519.Curve25519KeyPair
import org.whispersystems.curve25519.java.curve_sigs
import xyz.johnny.norntalk.security.Curve.Companion.getCurve


/**
 * Classe permettant d'effectuer des échanges de clés Diffie-Hellman à l'aide de l'algorithme
 * asymmétrique Curve25519.
 * Cette classe est utilisée pour obtenir les classes ChaCha correspondant à chaque conversation.
 * La méthode statique [getCurve] permet d'obtenir l'instance de la classe associée à l'utilisateur.
 *
 * @property secretKey Clé privée
 * @property publicKey Clé publique
 * @constructor Crée une instance de la classe Curve à l'aide d'une clé privée et d'une clé publique
 */
class Curve(private val secretKey: ByteArray, val publicKey: ByteArray) {

    companion object {

        /**
         * L'algorithme de chiffrement Curve25519
         */
        val cipher = Security.curve

        /**
         * Génère une paire de clés Curve25519
         *
         * @return Paire de clés
         */
        fun generateKeyPair(): Curve25519KeyPair = cipher.generateKeyPair()

        /**
         * Créé l'instance par défaut de la classe Curve et conserve la clé privée de l'application
         * de façon à pouvoir recréer la même instance lors d'un futur lancement de l'application
         *
         * @param context Contexte de l'application
         * @return Instance de Curve de l'utilisateur
         */
        fun getCurve(context: Context): Curve {
            // récupérer la clé privée de l'utilisateur si elle existe
            val settings = context.getSharedPreferences("CURVE", Context.MODE_PRIVATE)
            val pk = settings.getString("privateKey", null)
            val privateKey: ByteArray

            if (pk == null) {
                // générer la paire de clé si elle n'existe pas
                val kp = cipher.generateKeyPair()

                privateKey = kp.privateKey

                // conserver la clé privée de l'application en la codant en base64
                val editor = settings.edit()
                editor.putString("privateKey", Base64.encodeToString(privateKey, Base64.DEFAULT))
                editor.apply()
            } else {
                // si elle existe il faut la décoder
                privateKey = Base64.decode(pk, Base64.DEFAULT)
            }

            // générer la clé publique
            val publicKey = ByteArray(ChaCha.KEY_SIZE)
            curve_sigs.curve25519_keygen(publicKey, privateKey)

            // retourner l'instance de la classe Curve
            return Curve(privateKey, publicKey)
        }

    }

    /**
     * Créer une clé secrète partagée avec l'utilisateur possédant la clé publique otherPK
     *
     * @param otherPK Clé publique d'un autre utilisateur
     * @return Clé secrète
     */
    fun getCipher(otherPK: ByteArray) = getCipher(arrayOf(otherPK))

    /**
     * Créer une clé secrète partagée avec les différents utilisateur
     * possédant les clés publiques otherPKs
     *
     * @param otherPKs Clé publique d'un autre utilisateur
     * @return Clé secrète
     */
    fun getCipher(otherPKs: Array<ByteArray>): ChaCha {
        var secretKey = this.secretKey
        for (pk in otherPKs) {
            secretKey = cipher.calculateAgreement(pk, secretKey)
        }

        return ChaCha(secretKey)
    }

}
