package xyz.johnny.norntalk.security

import android.Manifest
import android.annotation.SuppressLint
import android.content.Context
import android.content.DialogInterface
import android.content.SharedPreferences
import android.content.pm.PackageManager
import android.graphics.Bitmap
import android.os.AsyncTask
import android.os.Handler
import android.os.Looper
import android.os.Message
import android.support.v4.content.ContextCompat
import android.support.v7.app.AlertDialog
import android.telephony.TelephonyManager
import android.util.Log
import android.view.WindowManager
import android.widget.EditText
import android.widget.Toast
import com.google.zxing.BarcodeFormat
import org.bouncycastle.util.encoders.Hex
import xyz.johnny.norntalk.R
import xyz.johnny.norntalk.messages.NornContact
import xyz.johnny.norntalk.messages.NornMessage
import java.lang.ref.WeakReference
import java.util.*

/**
 * Classe permettant de générer les QR codes
 * Les QR codes générés sont composés en 4 parties:
 *
 *  1. La clé publique de l'utilisateur
 *  2. Le numéro de téléphone de l'utilisateur
 *  3. Une clé à utilisation unique associée au QR code
 *  4. Un token (ou jeton) associé au QR code
 *
 * Cela permet d'envoyer la clé de l'utilisateur scannant le QR code par SMS en chiffrant sa clé
 * publique à l'aide de la clé à utilisation unique.
 * Il sera possible de reconnaitre ce SMS par la présence du token dans le message.
 * L'échange de clé ainsi effectué permet d'éviter les attaques de type man-in-the-middle.
 */
object QRCode {

    class CheckNumberTask(context: Context, val tmpNumber: String, val numberObj: NumberObject, val alertDialog: AlertDialog) : AsyncTask<Void, Void, Boolean>() {

        val context = WeakReference<Context>(context)

        override fun doInBackground(vararg params: Void?): Boolean {
            // vérifier le numéro de téléphone
            return CheckNumber.checkNumber(tmpNumber, this.context.get()!!)
        }

        override fun onPostExecute(result: Boolean) {
            super.onPostExecute(result)

            val context = this.context.get()!!

            if (result) {
                // transmettre le numéro de téléphone
                synchronized(numberObj) {
                    numberObj.number = NornContact.format(tmpNumber, context)
                    alertDialog.dismiss()
                    numberObj.notifyAll()
                }
                Toast.makeText(context, R.string.correct_number, Toast.LENGTH_SHORT).show()
            } else {
                Toast.makeText(context, R.string.wrong_number, Toast.LENGTH_SHORT).show()
            }
        }
    }

    /**
     * Taille du token présent dans les QR codes
     */
    val TOKEN_SIZE = 32

    /**
     * Préférences liées aux QR codes
     */
    private var preferences: SharedPreferences? = null

    /**
     * Association des tokens à leur clé à utilisation unique
     */
    val tempKeys = HashMap<String, ByteArray>()

    /**
     * Retourne les préférences relatives aux QR codes
     *
     * @param context Contexte courant
     * @return Préférences des QR codes
     */
    private fun getPreferences(context: Context): SharedPreferences {
        if (preferences == null)
            preferences = context.getSharedPreferences("QR", 0)

        return preferences!!
    }

    private var lastBitmap : Bitmap? = null

    /**
     * Crée une pop-up pour demander le numéro de téléphone à l'utilisateur
     *
     * @param context Contexte courant
     * @param numberObj Stocke le numéro de téléphone
     */
    private fun popUpNumber(context: Context, numberObj: NumberObject) {
        // Création de la popup
        val alertDialog = AlertDialog.Builder(context).create()
        alertDialog.setTitle(R.string.enter_number)

        // inclure la zone de texte dans la popup
        val inflater = alertDialog.layoutInflater
        val dialogView = inflater.inflate(R.layout.num_alert, null)
        alertDialog.setView(dialogView)

        // Affichage du clavier pour l'EditText
        alertDialog.window.setSoftInputMode(WindowManager.LayoutParams.SOFT_INPUT_STATE_VISIBLE)

        // Bouton 'OK'
        alertDialog.setButton(AlertDialog.BUTTON_POSITIVE, context.getString(R.string.OK), { _, _ -> })

        val cancel = { dialog: DialogInterface ->
            // transmettre un numéro vide en fermant la popup
            synchronized(numberObj) {
                numberObj.number = ""
                dialog.dismiss()
                numberObj.notifyAll()
            }
        }
        // Bouton 'Annuler' et action correspondante
        alertDialog.setButton(AlertDialog.BUTTON_NEGATIVE, context.getString(R.string.cancel), {
            dialogInterface, _ -> cancel(dialogInterface)
        })

        // action executée lorsque l'utilisateur appuie sur le bouton 'back'
        alertDialog.setOnCancelListener(cancel)

        alertDialog.setOnShowListener {
            alertDialog.getButton(AlertDialog.BUTTON_POSITIVE).setOnClickListener {
                // récupérer la valeur du champs de texte
                val tmpNumber = alertDialog.findViewById<EditText>(R.id.etNum)?.text.toString()

                CheckNumberTask(context, tmpNumber, numberObj, alertDialog).executeOnExecutor(AsyncTask.THREAD_POOL_EXECUTOR)
            }
        }

        alertDialog.show()
    }

    /**
     * Demande le numéro de téléphone à l'utilisateur
     */
    private fun getNumberFromUser(context: Context): String {
        val numberObj = NumberObject()

        synchronized (numberObj) {
            // crée la popup
            object : Handler(Looper.getMainLooper()) {

                override fun handleMessage(msg: Message?) {
                    popUpNumber(context, numberObj)
                }

            }.sendEmptyMessage(0)

            // attendre que la popup soit fermée
            while (numberObj.number == null)
                numberObj.wait()
        }

        return numberObj.number!!
    }

    /**
     * Trouve le numéro de téléphone ou, si ce n'est pas possible autrement, le demande à
     * l'utilisateur
     *
     * @param context Contexte courant
     * @param auto Si Vrai ne demandera pas le numéro à l'utilisateur si le numéro est introuvable
     * @return Numéro de téléphone de l'utilisateur
     */
    @SuppressLint("HardwareIds")
    fun getNumber(context: Context, auto: Boolean = false): String {
        if (CheckNumber.number.isNotEmpty())
            return CheckNumber.number

        Log.d(this::class.java.simpleName, "get phone number")
        val preferences = this.getPreferences(context)
        var number = preferences.getString("number", "")

        if (number.isEmpty()) {
            // essayer de trouver le numéro de téléphone directement si la permission a été donnée
            if (ContextCompat.checkSelfPermission(context, Manifest.permission.READ_PHONE_STATE) == PackageManager.PERMISSION_GRANTED) {
                // essayer de récupérer le numéro de téléphone de l'utilisateur
                val telephonyMgr = context.getSystemService(Context.TELEPHONY_SERVICE) as TelephonyManager
                val telMgrNumber = telephonyMgr.line1Number

                // vérifier si le numéro est correct et retourner le numéro si c'est le cas
                if (CheckNumber.checkNumber(telMgrNumber, context))
                    return telMgrNumber
            }

            // demander le numéro de téléphone à l'utilisateur
            number = if (auto) "" else getNumberFromUser(context)
        }

        // enregistrer le numéro de téléphone
        if (!number.isEmpty()) {
            val editor = preferences.edit()
            editor.putString("number", number)
            editor.apply()
        }

        Log.d(this::class.java.simpleName, "got number: " + CheckNumber.number)

        return number
    }

    /**
     * Génère un QR code contenant la clé publique de l'utilisateur, le numéro de téléphone de
     * l'utilisateur, une clé à utilisation unique et un token
     *
     * @param context Contexte courant
     * @return Bitmap du QR code ou null s'il est impossible de créer le QR code
     */
    fun generate(context: Context, auto: Boolean = false) : Bitmap? {

        // trouver le numéro de l'utilisateur
        val number = getNumber(context, auto)
        Log.d(this::class.java.simpleName, "Number: " + number)

        // retourner null si le numéro n'a pas pu être trouvé
        if (number.isEmpty())
            return null

        // récupérer la clé publique de l'utilisateur
        val pubkey = Curve.getCurve(context).publicKey

        // générer une clé à usage unique
        val tempkey = ByteArray(ChaCha.KEY_SIZE)
        Security.random.nextBytes(tempkey)

        // générer un token
        val btoken = ByteArray(TOKEN_SIZE)
        Security.random.nextBytes(btoken)
        val token = Hex.toHexString(btoken)

        Log.d(this::class.java.simpleName, token)

        // ajouter la clé et le jeton dans l'association
        this.tempKeys.put(token, tempkey)

        // concaténation des champs en une chaine de caractères
        val content = Hex.toHexString(pubkey) + ":" + number + ":" + Hex.toHexString(tempkey) + ":" + token

        // génération du QR code
        val QRencoder = Security.barcodeEncoder
        this.lastBitmap?.recycle()
        this.lastBitmap = QRencoder.encodeBitmap(content, BarcodeFormat.QR_CODE, 512, 512)
        return lastBitmap
    }

    /**
     * Méthode appelé à la réception de chaque message afin de vérifier si message reçu est
     * présent dans l'association jeton/clé.
     * Si c'est le cas, une conversation sécurisée sera créée avec l'expéditeur du message.
     *
     * @param context Contexte courant
     * @param msg SMS reçu
     * @return Vrai si le message contient un token existant faux sinon
     */
    fun message_hook(context: Context, msg: NornMessage): Boolean {
        // retourner faux si le message a une taille inférieure à une clé ChaCha20
        if (msg.rawtext.length <= ChaCha.KEY_SIZE)
            return false

        // récupérer le token, qui est suivi d'un ':' dans le message
        val token = msg.rawtext.substringBefore(':')

        // retourner faux si le token n'est pas dans l'association jeton/clé
        val tempKey = this.tempKeys[token] ?: return false

        // afficher un message pour signaler que le QR code a été utilisé
        Toast.makeText(context, context.getString(R.string.qr_used), Toast.LENGTH_SHORT).show()

        // déchiffrer la clé du correspondant à l'aide de la clé à usage unique
        val cipher = ChaCha(tempKey)
        val ciphertext = Hex.decode(msg.rawtext.substringAfter(':'))
        val key = cipher.decrypt(ciphertext)

        // associer la clé publique au contact
        msg.sender!!.setPublicKey(key, context)
        // démarrer la conversation sécurisée
        msg.conversation.setSecured(true, context)

        return true
    }

    /**
     * Classe permettant de décomposer le contenu d'un QR code créé par la méthode QRCode.generate
     *
     * @param content Contenu du QR code
     */
    class Decoder(content: String) {

        /**
         * clé publique de l'utilisateur ayant généré le QR code
         */
        val pubkey: ByteArray
        /**
         * numéro de téléphone de l'utilisateur ayant généré le QR code
         */
        val number: String
        /**
         * clé à usage unique à utiliser pour chiffrer la clé publique de l'utilisateur scannant le
         * QR code
         */
        val tempkey: ByteArray
        /**
         * token associé au QR code et à insérer au début du SMS
         */
        val token: String

        init {
            val parts = content.split(':')

            this.pubkey = Hex.decode(parts[0])
            this.number = parts[1]
            this.tempkey = Hex.decode(parts[2])
            this.token = parts[3]
        }

    }

}