package xyz.johnny.norntalk.ui_components

import android.content.Context
import android.content.DialogInterface
import android.content.Intent
import android.content.SharedPreferences
import android.os.Bundle
import android.preference.PreferenceManager
import android.provider.Telephony
import android.support.v7.app.AlertDialog
import android.support.v7.app.AppCompatActivity
import android.view.WindowManager
import android.widget.EditText
import android.widget.Switch
import android.widget.Toast
import kotlinx.android.synthetic.main.passalert_layout.*
import org.bouncycastle.util.encoders.Hex
import xyz.johnny.norntalk.R
import xyz.johnny.norntalk.SplashActivity
import xyz.johnny.norntalk.security.LoadHelper
import xyz.johnny.norntalk.security.QRCode
import xyz.johnny.norntalk.security.Security
import java.security.MessageDigest

/**
 * Classe activité de base de Norn Talk.
 * Gère la demande du mot de passe lors de l'accès à l'application si un mot de passe a été saisi
 * par l'utilisateur.
 */
abstract class BaseActivity: AppCompatActivity() {

    companion object {

        /**
         * Activité courante de l'application
         */
        var currentActivity: BaseActivity? = null

        /**
         * Algorithme de hachage utilisé
         */
        val HASH = "SHA-256"

        /**
         * Taille du vecteur d'initialisation du hash en octets
         */
        val HASH_IV_SIZE = 32

    }

    /**
     * Les classes héritant de [BaseActivity] écrasent cet attribut pour changer leur disposition
     */
    protected abstract val layout: Int

    private var _preferences: SharedPreferences? = null
    /**
     * Accès aux paramètres de l'application
     */
    protected val preferences get() = _preferences!!

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)

        this.setContentView(this.layout)

        PreferenceManager.setDefaultValues(this, "Default", Context.MODE_PRIVATE, R.xml.preferences, false)
        this._preferences = PreferenceManager.getDefaultSharedPreferences(this)

        this.initialize()
    }

    /**
     * Méthode appelée à la fin de onCreate
     */
    open fun initialize() {}

    override fun onStart() {
        super.onStart()
        // demander de définir Norn Talk comme application SMS par défaut
        if (Telephony.Sms.getDefaultSmsPackage(applicationContext) != applicationContext.packageName) {
            val intent = Intent(Telephony.Sms.Intents.ACTION_CHANGE_DEFAULT)
            intent.putExtra(Telephony.Sms.Intents.EXTRA_PACKAGE_NAME, applicationContext.packageName)
            startActivity(intent)
        }
    }

    override fun onResume() {
        // demander le mot de passe si l'utilisateur est sorti de l'application
        if (currentActivity == null && !NornCaptureActivity.scanning)
            if (this.preferences.getString(getString(R.string.password_key), getString(R.string.default_password)) != getString(R.string.default_password))
                popUpPasswd(null)

        // changer l'activité courante
        currentActivity = this

        if (Telephony.Sms.getDefaultSmsPackage(applicationContext) == applicationContext.packageName)
            LoadHelper { QRCode.getNumber(this, true) }

        super.onResume()
    }

    override fun onStop() {
        super.onStop()

        // donner la valeur null à currentActivity si l'utilisateur est sorti de l'application
        if (currentActivity == this)
            currentActivity = null
    }

    /**
     * calcul du hash du mot de passe
     *
     * @param password mot de passe entré par l'utilisateur
     * @return mot de passe haché
     */
    fun hash(password: String): String {
        val passwordHash = MessageDigest.getInstance(HASH)

        // lire le mot de passe de l'application
        val prefPass = this.preferences.getString(getString(R.string.password_key), getString(R.string.default_password))

        var iv = ByteArray(HASH_IV_SIZE)
        if (prefPass.isEmpty()) {
            // générer le vecteur d'initialisation si le mot de passe n'existe pas
            Security.random.nextBytes(iv)
        } else {
            // lire le vecteur d'initialisation si le message existe
            iv = Hex.decode(prefPass).copyOf(HASH_IV_SIZE)
        }

        // créer le hash
        passwordHash.update(iv)
        passwordHash.update(password.toByteArray(Charsets.UTF_8))
        return Hex.toHexString(iv + passwordHash.digest())
    }

    /**
     * Crée la popup demandant le mot de passe
     *
     * @param switchPasswd si null demande simplement le mot de passe,
     *                     sinon active le switch lorsque le mot de passe est validé par l'utilisateur
     */
    fun popUpPasswd(switchPasswd: Switch?) {
        val alertDialog = AlertDialog.Builder(this).create()
        alertDialog.setTitle(R.string.textChoixMDP)

        // inclure l'EditText dans la popup
        val inflater = this.layoutInflater
        val dialogView = inflater.inflate(R.layout.passalert_layout, password_group)
        alertDialog.setView(dialogView)

        // trouver le champs de saisie de mot de passe
        val etPasswd = dialogView.findViewById<EditText>(R.id.etPasswd)!!
        // Affichage du clavier pour l'EditText
        alertDialog.window.setSoftInputMode(WindowManager.LayoutParams.SOFT_INPUT_STATE_VISIBLE)

        // action correspondante au bouton OK
        val actionOk: (DialogInterface) -> Unit
        if (switchPasswd == null) {
            actionOk = { dialogInterface ->
                // vérifier le mot de passe saisi par l'utilisateur
                val password = this.hash(etPasswd.text.toString())

                // relancer la pop up si le mot de passe est incorrect
                if (password == this.preferences.getString(getString(R.string.password_key), getString(R.string.default_password))) {
                    dialogInterface.dismiss()
                } else {
                    Toast.makeText(this, R.string.wrong_pass, Toast.LENGTH_SHORT).show()
                }
            }
        } else {
            actionOk = { dialogInterface ->
                val password = this.hash(etPasswd.text.toString())

                //Enregistrer le mot de passe saisi
                val preferenceEditor = this.preferences.edit()
                preferenceEditor.putString(getString(R.string.password_key), password)
                preferenceEditor.apply()

                dialogInterface.dismiss()
            }
        }

        // action correspondante au bouton annuler
        val cancel: (DialogInterface) -> Unit
        if (switchPasswd == null)
            cancel = { System.exit(0) }
        else
            cancel = { dialogInterface ->
                switchPasswd.isChecked = false
                dialogInterface.cancel()
            }

        // lancer l'action annuler si l'utilisateur sort de la popup en appuyant sur le bouton retour
        alertDialog.setOnCancelListener(cancel)

        // créer le bouton OK
        alertDialog.setButton(AlertDialog.BUTTON_POSITIVE, getString(R.string.OK), { _, _ -> })

        // Afficher le bouton annuler uniquement si on demande le mot de passe pour la première fois
        if (switchPasswd != null) {
            alertDialog.setButton(AlertDialog.BUTTON_NEGATIVE, getString(R.string.cancel), {
                dialog, _ -> cancel(dialog)
            })
        }

        alertDialog.setOnShowListener({dialogInterface ->
            // création du bouton OK
            alertDialog.getButton(AlertDialog.BUTTON_POSITIVE).setOnClickListener { view ->
                actionOk(dialogInterface)
            }
        })

        //Affichage de la popup
        alertDialog.show()
    }

}